#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict

from ovos_number_parser.util import convert_to_mixed_fraction, is_numeric, look_for_fractions, \
    invert_dict, ReplaceableNumber, partition_list, tokenize, Token

_NUM_STRING_RU = {
    0: 'ноль',
    1: 'один',
    2: 'два',
    3: 'три',
    4: 'четыре',
    5: 'пять',
    6: 'шесть',
    7: 'семь',
    8: 'восемь',
    9: 'девять',
    10: 'десять',
    11: 'одиннадцать',
    12: 'двенадцать',
    13: 'тринадцать',
    14: 'четырнадцать',
    15: 'пятнадцать',
    16: 'шестнадцать',
    17: 'семнадцать',
    18: 'восемнадцать',
    19: 'девятнадцать',
    20: 'двадцать',
    30: 'тридцать',
    40: 'сорок',
    50: 'пятьдесят',
    60: 'шестьдесят',
    70: 'семьдесят',
    80: 'восемьдесят',
    90: 'девяносто',
    100: 'сто',
    200: 'двести',
    300: 'триста',
    400: 'четыреста',
    500: 'пятьсот',
    600: 'шестьсот',
    700: 'семьсот',
    800: 'восемьсот',
    900: 'девятьсот'
}

_FRACTION_STRING_RU = {
    2: 'половина',
    3: 'треть',
    4: 'четверть',
    5: 'пятая',
    6: 'шестая',
    7: 'седьмая',
    8: 'восьмая',
    9: 'девятая',
    10: 'десятая',
    11: 'одиннадцатая',
    12: 'двенадцатая',
    13: 'тринадцатая',
    14: 'четырнадцатая',
    15: 'пятнадцатая',
    16: 'шестнадцатая',
    17: 'семнадцатая',
    18: 'восемнадцатая',
    19: 'девятнадцатая',
    20: 'двадцатая',
    30: 'тридцатая',
    40: 'сороковая',
    50: 'пятидесятая',
    60: 'шестидесятая',
    70: 'семидесятая',
    80: 'восьмидесятая',
    90: 'девяностая',
    1e2: 'сотая',
    1e3: 'тысячная',
    1e6: 'миллионная',
    1e9: 'миллиардная'
}

_SHORT_SCALE_RU = OrderedDict([
    (1e3, 'тысяча'),
    (1e6, "миллион"),
    (1e9, "миллиард"),
    (1e12, "триллион"),
    (1e15, "квадриллион"),
    (1e18, "квинтиллион"),
    (1e21, "секстиллион"),
    (1e24, "септиллион"),
    (1e27, "октиллион"),
    (1e30, "нониллион"),
    (1e33, "дециллион"),
    (1e36, "ундециллион"),
    (1e39, "дуодециллион"),
    (1e42, "тредециллион"),
    (1e45, "кваттордециллион"),
    (1e48, "квиндециллион"),
    (1e51, "сексдециллион"),
    (1e54, "септендециллион"),
    (1e57, "октодециллион"),
    (1e60, "новемдециллион"),
    (1e63, "вигинтиллион"),
    (1e66, "унвигинтиллион"),
    (1e69, "дуовигинтиллион"),
    (1e72, "тревигинтиллион"),
    (1e75, "кватторвигинтиллион"),
    (1e78, "квинвигинтиллион"),
    (1e81, "секснвигинтиллион"),
    (1e84, "септенвигинтиллион"),
    (1e87, "октовигинтиллион"),
    (1e90, "новемвигинтиллион"),
    (1e93, "тригинтиллион"),
])

_LONG_SCALE_RU = OrderedDict([
    (1e3, 'тысяча'),
    (1e6, "миллион"),
    (1e9, "миллиард"),
    (1e12, "биллион"),
    (1e15, "биллиард"),
    (1e18, "триллион"),
    (1e21, "триллиард"),
    (1e24, "квадриллион"),
    (1e27, "квадриллиард"),
    (1e30, "квинтиллион"),
    (1e33, "квинтиллиард"),
    (1e36, "секстиллион"),
    (1e39, "секстиллиард"),
    (1e42, "септиллион"),
    (1e45, "септиллиард"),
    (1e48, "октиллион"),
    (1e51, "октиллиард"),
    (1e54, "нониллион"),
    (1e57, "нониллиард"),
    (1e60, "дециллион"),
    (1e63, "дециллиард"),
    (1e66, "ундециллион"),
    (1e72, "дуодециллион"),
    (1e78, "тредециллион"),
    (1e84, "кваттордециллион"),
    (1e90, "квиндециллион"),
    (1e96, "сексдециллион"),
    (1e102, "септендециллион"),
    (1e108, "октодециллион"),
    (1e114, "новемдециллион"),
    (1e120, "вигинтиллион"),
])

_ORDINAL_BASE_RU = {
    1: 'первый',
    2: 'второй',
    3: 'третий',
    4: 'четвёртый',
    5: 'пятый',
    6: 'шестой',
    7: 'седьмой',
    8: 'восьмой',
    9: 'девятый',
    10: 'десятый',
    11: 'одиннадцатый',
    12: 'двенадцатый',
    13: 'тринадцатый',
    14: 'четырнадцатый',
    15: 'пятнадцатый',
    16: 'шестнадцатый',
    17: 'семнадцатый',
    18: 'восемнадцатый',
    19: 'девятнадцатый',
    20: 'двадцатый',
    30: 'тридцатый',
    40: "сороковой",
    50: "пятидесятый",
    60: "шестидесятый",
    70: "семидесятый",
    80: "восьмидесятый",
    90: "девяностый",
    1e2: "сотый",
    2e2: "двухсотый",
    3e2: "трёхсотый",
    4e2: "четырёхсотый",
    5e2: "пятисотый",
    6e2: "шестисотый",
    7e2: "семисотый",
    8e2: "восьмисотый",
    9e2: "девятисотый",
    1e3: "тысячный"
}

_SHORT_ORDINAL_RU = {
    1e6: "миллион",
    1e9: "миллиард",
    1e12: "триллион",
    1e15: "квадриллион",
    1e18: "квинтиллион",
    1e21: "секстиллион",
    1e24: "септиллион",
    1e27: "октиллион",
    1e30: "нониллион",
    1e33: "дециллион",
    1e36: "ундециллион",
    1e39: "дуодециллион",
    1e42: "тредециллион",
    1e45: "кваттордециллион",
    1e48: "квиндециллион",
    1e51: "сексдециллион",
    1e54: "септендециллион",
    1e57: "октодециллион",
    1e60: "новемдециллион",
    1e63: "вигинтиллион"
}
_SHORT_ORDINAL_RU.update(_ORDINAL_BASE_RU)

_LONG_ORDINAL_RU = {
    1e6: "миллион",
    1e9: "миллиард",
    1e12: "биллион",
    1e15: "биллиард",
    1e18: "триллион",
    1e21: "триллиард",
    1e24: "квадриллион",
    1e27: "квадриллиард",
    1e30: "квинтиллион",
    1e33: "квинтиллиард",
    1e36: "секстиллион",
    1e39: "секстиллиард",
    1e42: "септиллион",
    1e45: "септиллиард",
    1e48: "октиллион",
    1e51: "октиллиард",
    1e54: "нониллион",
    1e57: "нониллиард",
    1e60: "дециллион",
    1e63: "дециллиард",
    1e66: "ундециллион",
    1e72: "дуодециллион",
    1e78: "тредециллион",
    1e84: "кваттордециллион",
    1e90: "квиндециллион",
    1e96: "сексдециллион",
    1e102: "септендециллион",
    1e108: "октодециллион",
    1e114: "новемдециллион",
    1e120: "вигинтиллион"
}
_LONG_ORDINAL_RU.update(_ORDINAL_BASE_RU)


def generate_plurals_ru(originals):
    """
    Return a new set or dict containing the plural form of the original values,

    In English this means all with 's' appended to them.

    Args:
        originals set(str) or dict(str, any): values to pluralize

    Returns:
        set(str) or dict(str, any)

    """
    suffixes = ["а", "ах", "ам", "ами", "ные", "ный", "ов", "ом", "ы"]
    if isinstance(originals, dict):
        return {key + suffix: value for key, value in originals.items() for suffix in suffixes}
    return {value + suffix for value in originals for suffix in suffixes}


# negate next number (-2 = 0 - 2)
_NEGATIVES = {"минус"}

# sum the next number (twenty two = 20 + 2)
_SUMS = {'двадцать', '20', 'тридцать', '30', 'сорок', '40', 'пятьдесят', '50',
         'шестьдесят', '60', 'семьдесят', '70', 'восемьдесят', '80', 'девяносто', '90',
         'сто', '100', 'двести', '200', 'триста', '300', 'четыреста', '400',
         'пятьсот', '500', 'шестьсот', '600', 'семьсот', '700', 'восемьсот', '800',
         'девятьсот', '900'}

_MULTIPLIES_LONG_SCALE_RU = set(_LONG_SCALE_RU.values()) | \
                            generate_plurals_ru(_LONG_SCALE_RU.values())

_MULTIPLIES_SHORT_SCALE_RU = set(_SHORT_SCALE_RU.values()) | \
                             generate_plurals_ru(_SHORT_SCALE_RU.values())

# split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
_FRACTION_MARKER = {"и", "с", " "}

# decimal marker ( 1 point 5 = 1 + 0.5)
_DECIMAL_MARKER = {"целая", "целых", "точка", "запятая"}

_STRING_NUM_RU = invert_dict(_NUM_STRING_RU)
_STRING_NUM_RU.update({
    "тысяч": 1e3,
})
_STRING_NUM_RU.update(generate_plurals_ru(_STRING_NUM_RU))
_STRING_NUM_RU.update({
    "четверти": 0.25,
    "четвёртая": 0.25,
    "четвёртых": 0.25,
    "третья": 1 / 3,
    "третяя": 1 / 3,
    "вторая": 0.5,
    "вторых": 0.5,
    "половина": 0.5,
    "половиной": 0.5,
    "пол": 0.5,
    "одна": 1,
    "двойка": 2,
    "двое": 2,
    "пара": 2,
    "сот": 100,
    "сотен": 100,
    "сотни": 100,
    "сотня": 100,
})

_WORDS_NEXT_RU = [
    "будущая", "будущее", "будущей", "будущий", "будущим", "будущую",
    "новая", "новое", "новой", "новый", "новым",
    "следующая", "следующее", "следующей", "следующем", "следующий", "следующую",
]
_WORDS_PREV_RU = [
    "предыдущая", "предыдущем", "предыдущей", "предыдущий", "предыдущим", "предыдущую",
    "прошедшая", "прошедшем", "прошедшей", "прошедший", "прошедшим", "прошедшую",
    "прошлая", "прошлой", "прошлом", "прошлую", "прошлый", "прошлым",
    "том", "тот",
]
_WORDS_CURRENT_RU = [
    "данная", "данное", "данном", "данный",
    "настойщая", "настоящее", "настойщем", "настойщем", "настойщий",
    "нынешняя", "нынешнее", "нынешней", "нынешнем", "нынешний",
    "текущая", "текущее", "текущей", "текущем", "текущий",
    "это", "этим", "этой", "этом", "этот", "эту",
]
_WORDS_NOW_RU = [
    "теперь",
    "сейчас",
]
_WORDS_MORNING_RU = ["утро", "утром"]
_WORDS_DAY_RU = ["днём"]
_WORDS_EVENING_RU = ["вечер", "вечером"]
_WORDS_NIGHT_RU = ["ночь", "ночью"]

_STRING_SHORT_ORDINAL_RU = invert_dict(_SHORT_ORDINAL_RU)
_STRING_LONG_ORDINAL_RU = invert_dict(_LONG_ORDINAL_RU)


def nice_number_ru(number, speech=True, denominators=range(1, 21)):
    """ English helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 and a half" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_RU[den]
    if whole == 0:
        if num == 1 and den <= 4:
            return_string = '{}'.format(den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1 and den == 2:
        return_string = '{} с половиной'.format(whole)
    else:
        return_string = '{} и {} {}'.format(whole, num, den_str)
    if 2 <= den <= 4:
        if 2 <= num <= 4:
            return_string = return_string[:-1] + 'и'
        elif num > 4:
            return_string = return_string[:-1] + 'ей'
    elif den >= 5:
        if 2 <= num <= 4:
            return_string = return_string[:-2] + 'ые'
        elif num > 4:
            return_string = return_string[:-2] + 'ых'

    return return_string


def pronounce_number_ru(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    num = number
    # deal with infinity
    if num == float("inf"):
        return "бесконечность"
    elif num == float("-inf"):
        return "минус бесконечность"
    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            if ordinals:
                # This handles negative powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} на десять в {}{} степени'.format(
                    'минус ' if float(n) < 0 else '',
                    pronounce_number_ru(
                        abs(float(n)), places, short_scale, False, ordinals=True),
                    'минус ' if power < 0 else '',
                    pronounce_number_ru(abs(power), places, short_scale, False, ordinals=True))
            else:
                # This handles negative powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} на десять в степени {}{}'.format(
                    'минус ' if float(n) < 0 else '',
                    pronounce_number_ru(
                        abs(float(n)), places, short_scale, False, ordinals=False),
                    'минус ' if power < 0 else '',
                    pronounce_number_ru(abs(power), places, short_scale, False, ordinals=False))

    if short_scale:
        number_names = _NUM_STRING_RU.copy()
        number_names.update(_SHORT_SCALE_RU)
    else:
        number_names = _NUM_STRING_RU.copy()
        number_names.update(_LONG_SCALE_RU)

    digits = [number_names[n] for n in range(0, 20)]

    tens = [number_names[n] for n in range(10, 100, 10)]

    if short_scale:
        hundreds = [_SHORT_SCALE_RU[n] for n in _SHORT_SCALE_RU.keys()]
    else:
        hundreds = [_LONG_SCALE_RU[n] for n in _LONG_SCALE_RU.keys()]

    # deal with negative numbers
    result = ""
    if num < 0:
        result = "минус "
    num = abs(num)

    # check for a direct match
    if num in number_names and not ordinals:
        result += number_names[num]
    else:
        def _sub_thousand(n, ordinals=False):
            assert 0 <= n <= 999
            if n in _SHORT_ORDINAL_RU and ordinals:
                return _SHORT_ORDINAL_RU[n]
            if n <= 19:
                return digits[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                return tens[q - 1] + (" " + _sub_thousand(r, ordinals) if r
                                      else "")
            else:
                q, r = divmod(n, 100)
                return _NUM_STRING_RU[q * 100] + (" " + _sub_thousand(r, ordinals) if r else "")

        def _short_scale(n):
            if n > max(_SHORT_SCALE_RU.keys()):
                return "бесконечность"
            ordi = ordinals

            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000)):
                if not z:
                    continue
                number = _sub_thousand(z, not i and ordi)

                if i:
                    if i >= len(hundreds):
                        return ""
                    if ordi:
                        if i * 1000 in _SHORT_ORDINAL_RU:
                            if z == 1:
                                number = _SHORT_ORDINAL_RU[i * 1000]
                            else:
                                if z > 5:
                                    number = number[:-1] + "и"
                                number += _SHORT_ORDINAL_RU[i * 1000]
                        else:
                            if n not in _SHORT_SCALE_RU:
                                num = int("1" + "0" * (len(str(n)) // 3 * 3))

                                if number[-3:] == "два":
                                    number = number[:-1] + "ух"
                                elif number[-2:] == "ри" or number[-2:] == "ре":
                                    number = number[:-1] + "ёх"
                                elif number[-1:] == "ь":
                                    number = number[:-1] + "и"

                                number += _SHORT_SCALE_RU[num] + "ный"
                            else:
                                number = _SHORT_SCALE_RU[n] + "ный"
                    elif z == 1:
                        number = hundreds[i - 1]
                    else:
                        if i == 1:
                            if z % 10 == 1 and z % 100 // 10 != 1:
                                number = number[:-2] + "на"
                            elif z % 10 == 2 and z % 100 // 10 != 1:
                                number = number[:-1] + "е"
                            number += " " + plural_ru(z, "тысяча", "тысячи", "тысяч")
                        elif 1 <= z % 10 <= 4 and z % 100 // 10 != 1:
                            number += " " + hundreds[i - 1] + "а"
                        else:
                            number += " " + hundreds[i - 1] + "ов"

                res.append(number)
                ordi = False

            return " ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        def _long_scale(n):
            if n >= max(_LONG_SCALE_RU.keys()):
                return "бесконечность"
            ordi = ordinals
            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000000)):
                if not z:
                    continue
                number = pronounce_number_ru(z, places, True, scientific,
                                             ordinals=ordi and not i)
                # strip off the comma after the thousand
                if i:
                    if i >= len(hundreds):
                        return ""
                    # plus one as we skip 'thousand'
                    # (and 'hundred', but this is excluded by index value)
                    number = number.replace(',', '')

                    if ordi:
                        if (i + 1) * 1000000 in _LONG_ORDINAL_RU:
                            if z == 1:
                                number = _LONG_ORDINAL_RU[
                                    (i + 1) * 1000000]
                            else:
                                number += _LONG_ORDINAL_RU[
                                    (i + 1) * 1000000]
                        else:
                            if n not in _LONG_SCALE_RU:
                                num = int("1" + "0" * (len(str(n)) // 3 * 3))

                                if number[-3:] == "два":
                                    number = number[:-1] + "ух"
                                elif number[-2:] == "ри" or number[-2:] == "ре":
                                    number = number[:-1] + "ёх"
                                elif number[-1:] == "ь":
                                    number = number[:-1] + "и"

                                number += _LONG_SCALE_RU[num] + "ный"
                            else:
                                number = " " + _LONG_SCALE_RU[n] + "ный"
                    elif z == 1:
                        number = hundreds[i]
                    elif z <= 4:
                        number += " " + hundreds[i] + "а"
                    else:
                        number += " " + hundreds[i] + "ов"

                res.append(number)
            return " ".join(reversed(res))

        if short_scale:
            result += _short_scale(num)
        else:
            result += _long_scale(num)

    # deal with scientific notation unpronounceable as number
    if not result and "e" in str(num):
        return pronounce_number_ru(num, places, short_scale, scientific=True)
    # Deal with fractional part
    elif not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "минус " or not result):
            result += "ноль"
        result += " точка"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result


def plural_ru(num: int, one: str, few: str, many: str):
    num %= 100
    if num // 10 == 1:
        return many
    if num % 10 == 1:
        return one
    if 2 <= num % 10 <= 4:
        return few
    return many


def numbers_to_digits_ru(text, short_scale=True, ordinals=False):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numbers should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)

    Returns:
        str
        The original text, with numbers subbed in where appropriate.

    """
    text = text.lower()
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_ru(tokens, short_scale, ordinals)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_ru(tokens, short_scale=True,
                                  ordinals=False, fractional_numbers=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_ru(tokens, short_scale,
                                         ordinals, fractional_numbers)

        if not to_replace:
            break

        results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_ru(tokens, short_scale=True,
                                 ordinals=False, fractional_numbers=True):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = \
        _extract_number_with_text_ru_helper(tokens, short_scale,
                                            ordinals, fractional_numbers)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_ru_helper(tokens,
                                        short_scale=True, ordinals=False,
                                        fractional_numbers=True):
    """
    Helper for _extract_number_with_text_en.

    This contains the real logic for parsing, but produces
    a result that needs a little cleaning (specific, it may
    contain leading articles that can be trimmed off).

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:

    Returns:
        int or float, [Tokens]

    """
    if fractional_numbers:
        fraction, fraction_text = \
            _extract_fraction_with_text_ru(tokens, short_scale, ordinals)
        if fraction:
            return fraction, fraction_text

        decimal, decimal_text = \
            _extract_decimal_with_text_ru(tokens, short_scale, ordinals)
        if decimal:
            return decimal, decimal_text

    return _extract_whole_number_with_text_ru(tokens, short_scale, ordinals)


def _extract_fraction_with_text_ru(tokens, short_scale, ordinals):
    """
    Extract fraction numbers from a string.

    This function handles text such as '2 and 3/4'. Note that "one half" or
    similar will be parsed by the whole number function.

    Args:
        tokens [Token]: words and their indexes in the original string.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (int or float, [Token])
        The value found, and the list of relevant tokens.
        (None, None) if no fraction value is found.

    """
    for c in _FRACTION_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_ru(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_ru(partitions[2], short_scale,
                                              ordinals, fractional_numbers=True)

            if not numbers1 or not numbers2:
                return None, None

            # ensure first is not a fraction and second is a fraction
            num1 = numbers1[-1]
            num2 = numbers2[0]
            if num1.value >= 1 and 0 < num2.value < 1:
                return num1.value + num2.value, \
                       num1.tokens + partitions[1] + num2.tokens

    return None, None


def _extract_decimal_with_text_ru(tokens, short_scale, ordinals):
    """
    Extract decimal numbers from a string.

    This function handles text such as '2 point 5'.

    Notes:
        While this is a helper for extract_number_xx, it also depends on
        extract_number_xx, to parse out the components of the decimal.

        This does not currently handle things like:
            number dot number number number

    Args:
        tokens [Token]: The text to parse.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (float, [Token])
        The value found and relevant tokens.
        (None, None) if no decimal value is found.

    """
    for c in _DECIMAL_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_ru(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_ru(partitions[2], short_scale,
                                              ordinals, fractional_numbers=False)

            if not numbers1 or not numbers2:
                return None, None

            number = numbers1[-1]
            decimal = numbers2[0]

            # TODO handle number dot number number number
            if "." not in str(decimal.text):
                return number.value + float('0.' + str(decimal.value)), \
                       number.tokens + partitions[1] + decimal.tokens
    return None, None


def _extract_whole_number_with_text_ru(tokens, short_scale, ordinals):
    """
    Handle numbers not handled by the decimal or fraction functions. This is
    generally whole numbers. Note that phrases such as "one half" will be
    handled by this function, while "one and a half" are handled by the
    fraction function.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:

    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.

    """
    multiplies, string_num_ordinal, string_num_scale = \
        _initialize_number_data(short_scale)

    number_words = []  # type: [Token]
    val = False
    prev_val = None
    next_val = None
    to_sum = []
    for idx, token in enumerate(tokens):
        current_val = None
        if next_val:
            next_val = None
            continue

        word = token.word
        if word in word in _NEGATIVES:
            number_words.append(token)
            continue

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""

        # In Russian (?) we do no use suffix (1st,2nd,..) but use point instead (1.,2.,..)
        if is_numeric(word[:-1]) and \
                (word.endswith(".")):
            # explicit ordinals, 1st, 2nd, 3rd, 4th.... Nth
            word = word[:-1]

            # handle nth one
        #    if next_word == "one":
        # would return 1 instead otherwise
        #        tokens[idx + 1] = Token("", idx)
        #        next_word = ""

        # Normalize Russian inflection of numbers (один, одна, одно,...)
        if not ordinals:
            word = _text_ru_inflection_normalize(word, 1)

        if word not in string_num_scale and \
                word not in _STRING_NUM_RU and \
                word not in _SUMS and \
                word not in multiplies and \
                not (ordinals and word in string_num_ordinal) and \
                not is_numeric(word) and \
                not is_fractional_ru(word, short_scale=short_scale) and \
                not look_for_fractions(word.split('/')):
            words_only = [token.word for token in number_words]
            if number_words and not all([w in _NEGATIVES for w in words_only]):
                break
            else:
                number_words = []
                continue
        elif word not in multiplies \
                and prev_word not in multiplies \
                and prev_word not in _SUMS \
                and not (ordinals and prev_word in string_num_ordinal) \
                and prev_word not in _NEGATIVES:
            number_words = [token]
        elif prev_word in _SUMS and word in _SUMS:
            number_words = [token]
        else:
            number_words.append(token)

        # is this word already a number ?
        if is_numeric(word):
            if word.isdigit():  # doesn't work with decimals
                val = int(word)
            else:
                val = float(word)
            current_val = val

        # is this word the name of a number ?
        if word in _STRING_NUM_RU:
            val = _STRING_NUM_RU.get(word)
            current_val = val
        elif word in string_num_scale:
            val = string_num_scale.get(word)
            current_val = val
        elif ordinals and word in string_num_ordinal:
            val = string_num_ordinal[word]
            current_val = val

        # is the prev word an ordinal number and current word is one?
        # second one, third one
        if ordinals and prev_word in string_num_ordinal and val == 1:
            val = prev_val

        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if (prev_word in _SUMS and val and val < 10) \
                or (prev_word in _SUMS and val and val < 100 and prev_val >= 100) \
                or all([prev_word in multiplies, val < prev_val if prev_val else False]):
            val = prev_val + val

        # is the prev word a number and should we multiply it?
        # twenty hundred, six hundred
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val

        # is this a spoken fraction?
        # half cup
        if val is False:
            val = is_fractional_ru(word, short_scale=short_scale)
            current_val = val

        # 2 fifths
        if not ordinals:
            next_val = is_fractional_ru(next_word, short_scale=short_scale)
            if next_val:
                if not val:
                    val = 1
                val = val * next_val
                number_words.append(tokens[idx + 1])

        # is this a negative number?
        if val and prev_word and prev_word in _NEGATIVES:
            val = 0 - val

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            a_pieces = word.split('/')
            if look_for_fractions(a_pieces):
                val = float(a_pieces[0]) / float(a_pieces[1])
        else:
            if all([
                prev_word in _SUMS,
                word not in _SUMS,
                word not in multiplies,
                current_val >= 10
            ]):
                # Backtrack - we've got numbers we can't sum.
                number_words.pop()
                val = prev_val
                break
            prev_val = val

            if word in multiplies and next_word not in multiplies:
                # handle long numbers
                # six hundred sixty six
                # two million five hundred thousand
                #
                # This logic is somewhat complex, and warrants
                # extensive documentation for the next coder's sake.
                #
                # The current word is a power of ten. `current_val` is
                # its integer value. `val` is our working sum
                # (above, when `current_val` is 1 million, `val` is
                # 2 million.)
                #
                # We have a dict `string_num_scale` containing [value, word]
                # pairs for "all" powers of ten: string_num_scale[10] == "ten.
                #
                # We need go over the rest of the tokens, looking for other
                # powers of ten. If we find one, we compare it with the current
                # value, to see if it's smaller than the current power of ten.
                #
                # Numbers which are not powers of ten will be passed over.
                #
                # If all the remaining powers of ten are smaller than our
                # current value, we can set the current value aside for later,
                # and begin extracting another portion of our final result.
                # For example, suppose we have the following string.
                # The current word is "million".`val` is 9000000.
                # `current_val` is 1000000.
                #
                #    "nine **million** nine *hundred* seven **thousand**
                #     six *hundred* fifty seven"
                #
                # Iterating over the rest of the string, the current
                # value is larger than all remaining powers of ten.
                #
                # The if statement passes, and nine million (9000000)
                # is appended to `to_sum`.
                #
                # The main variables are reset, and the main loop begins
                # assembling another number, which will also be appended
                # under the same conditions.
                #
                # By the end of the main loop, to_sum will be a list of each
                # "place" from 100 up: [9000000, 907000, 600]
                #
                # The final three digits will be added to the sum of that list
                # at the end of the main loop, to produce the extracted number:
                #
                #    sum([9000000, 907000, 600]) + 57
                # == 9,000,000 + 907,000 + 600 + 57
                # == 9,907,657
                #
                # >>> foo = "nine million nine hundred seven thousand six
                #            hundred fifty seven"
                # >>> extract_number(foo)
                # 9907657

                time_to_sum = True
                for other_token in tokens[idx + 1:]:
                    if other_token.word in multiplies:
                        if string_num_scale[other_token.word] >= current_val:
                            time_to_sum = False
                        else:
                            continue
                    if not time_to_sum:
                        break
                if time_to_sum:
                    to_sum.append(val)
                    val = 0
                    prev_val = 0

    if val is not None and to_sum:
        val += sum(to_sum)

    return val, number_words


def _initialize_number_data(short_scale):
    """
    Generate dictionaries of words to numbers, based on scale.

    This is a helper function for _extract_whole_number.

    Args:
        short_scale boolean:

    Returns:
        (set(str), dict(str, number), dict(str, number))
        multiplies, string_num_ordinal, string_num_scale

    """
    multiplies = _MULTIPLIES_SHORT_SCALE_RU if short_scale \
        else _MULTIPLIES_LONG_SCALE_RU

    string_num_ordinal_ru = _STRING_SHORT_ORDINAL_RU if short_scale \
        else _STRING_LONG_ORDINAL_RU

    string_num_scale_ru = _SHORT_SCALE_RU if short_scale else _LONG_SCALE_RU
    string_num_scale_ru = invert_dict(string_num_scale_ru)
    string_num_scale_ru.update(generate_plurals_ru(string_num_scale_ru))
    return multiplies, string_num_ordinal_ru, string_num_scale_ru


def extract_number_ru(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    return _extract_number_with_text_ru(tokenize(text.lower()),
                                        short_scale, ordinals).value


def is_fractional_ru(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str[-3:] in ["тые", "тых"]:  # leading number is bigger than one (две четвёртые, три пятых)
        input_str = input_str[-3:] + "тая"
    fractions = {"целая": 1}  # first four numbers have little different format

    for num in _FRACTION_STRING_RU:  # Numbers from 2 to 1 hundred, more is not usually used in common speech
        if num > 1:
            fractions[_FRACTION_STRING_RU[num]] = num

    if input_str.lower() in fractions:
        return 1.0 / fractions[input_str.lower()]
    return False


def _text_ru_inflection_normalize(word, arg):
    """
    Russian Inflection normalizer.

    This try to normalize known inflection. This function is called
    from multiple places, each one is defined with arg.

    Args:
        word [Word]
        arg [Int]

    Returns:
        word [Word]

    """
    if word in ["тысяч", "тысячи"]:
        return "тысяча"

    if arg == 1:  # _extract_whole_number_with_text_ru
        if word in ["одна", "одним", "одно", "одной"]:
            return "один"
        if word == "две":
            return "два"
        if word == "пару":
            return "пара"

    elif arg == 2:  # extract_datetime_ru
        if word in ["часа", "часам", "часами", "часов", "часу"]:
            return "час"
        if word in ["минут", "минутам", "минутами", "минуту", "минуты"]:
            return "минута"
        if word in ["секунд", "секундам", "секундами", "секунду", "секунды"]:
            return "секунда"
        if word in ["дней", "дни"]:
            return "день"
        if word in ["неделе", "недели", "недель"]:
            return "неделя"
        if word in ["месяца", "месяцев"]:
            return "месяц"
        if word in ["года", "лет"]:
            return "год"
        if word in _WORDS_MORNING_RU:
            return "утром"
        if word in ["полудне", "полудня"]:
            return "полдень"
        if word in _WORDS_EVENING_RU:
            return "вечером"
        if word in _WORDS_NIGHT_RU:
            return "ночь"
        if word in ["викенд", "выходным", "выходных"]:
            return "выходные"
        if word in ["столетие", "столетий", "столетия"]:
            return "век"

        # Week days
        if word in ["среду", "среды"]:
            return "среда"
        if word in ["пятницу", "пятницы"]:
            return "пятница"
        if word in ["субботу", "субботы"]:
            return "суббота"

        # Months
        if word in ["марта", "марте"]:
            return "март"
        if word in ["мае", "мая"]:
            return "май"
        if word in ["августа", "августе"]:
            return "август"
    return word



from ovos_number_parser.util import convert_to_mixed_fraction, is_numeric, look_for_fractions

# Undefined articles ["un", "une"] cannot be supressed,
# in French, "un cheval" means "a horse" or "one horse".
_ARTICLES_FR = ["le", "la", "du", "de", "les", "des"]

_NUMBERS_FR = {
    "zéro": 0,
    "un": 1,
    "une": 1,
    "deux": 2,
    "trois": 3,
    "quatre": 4,
    "cinq": 5,
    "six": 6,
    "sept": 7,
    "huit": 8,
    "neuf": 9,
    "dix": 10,
    "onze": 11,
    "douze": 12,
    "treize": 13,
    "quatorze": 14,
    "quinze": 15,
    "seize": 16,
    "vingt": 20,
    "trente": 30,
    "quarante": 40,
    "cinquante": 50,
    "soixante": 60,
    "soixante-dix": 70,
    "septante": 70,
    "quatre-vingt": 80,
    "quatre-vingts": 80,
    "octante": 80,
    "huitante": 80,
    "quatre-vingt-dix": 90,
    "nonante": 90,
    "cent": 100,
    "cents": 100,
    "mille": 1000,
    "mil": 1000,
    "millier": 1000,
    "milliers": 1000,
    "million": 1000000,
    "millions": 1000000,
    "milliard": 1000000000,
    "milliards": 1000000000}

_ORDINAL_ENDINGS_FR = ("er", "re", "ère", "nd", "nde" "ième", "ème", "e")

_NUM_STRING_FR = {
    0: 'zéro',
    1: 'un',
    2: 'deux',
    3: 'trois',
    4: 'quatre',
    5: 'cinq',
    6: 'six',
    7: 'sept',
    8: 'huit',
    9: 'neuf',
    10: 'dix',
    11: 'onze',
    12: 'douze',
    13: 'treize',
    14: 'quatorze',
    15: 'quinze',
    16: 'seize',
    20: 'vingt',
    30: 'trente',
    40: 'quarante',
    50: 'cinquante',
    60: 'soixante',
    70: 'soixante-dix',
    80: 'quatre-vingt',
    90: 'quatre-vingt-dix'
}

_FRACTION_STRING_FR = {
    2: 'demi',
    3: 'tiers',
    4: 'quart',
    5: 'cinquième',
    6: 'sixième',
    7: 'septième',
    8: 'huitième',
    9: 'neuvième',
    10: 'dixième',
    11: 'onzième',
    12: 'douzième',
    13: 'treizième',
    14: 'quatorzième',
    15: 'quinzième',
    16: 'seizième',
    17: 'dix-septième',
    18: 'dix-huitième',
    19: 'dix-neuvième',
    20: 'vingtième'
}


def nice_number_fr(number, speech=True, denominators=range(1, 21)):
    """ French helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 et demi" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    strNumber = ""
    whole = 0
    num = 0
    den = 0

    result = convert_to_mixed_fraction(number, denominators)

    if not result:
        # Give up, just represent as a 3 decimal number
        whole = round(number, 3)
    else:
        whole, num, den = result

    if not speech:
        if num == 0:
            strNumber = '{:,}'.format(whole)
            strNumber = strNumber.replace(",", " ")
            strNumber = strNumber.replace(".", ",")
            return strNumber
        else:
            return '{} {}/{}'.format(whole, num, den)
    else:
        if num == 0:
            # if the number is not a fraction, nothing to do
            strNumber = str(whole)
            strNumber = strNumber.replace(".", ",")
            return strNumber
        den_str = _FRACTION_STRING_FR[den]
        # if it is not an integer
        if whole == 0:
            # if there is no whole number
            if num == 1:
                # if numerator is 1, return "un demi", for example
                strNumber = 'un {}'.format(den_str)
            else:
                # else return "quatre tiers", for example
                strNumber = '{} {}'.format(num, den_str)
        elif num == 1:
            # if there is a whole number and numerator is 1
            if den == 2:
                # if denominator is 2, return "1 et demi", for example
                strNumber = '{} et {}'.format(whole, den_str)
            else:
                # else return "1 et 1 tiers", for example
                strNumber = '{} et 1 {}'.format(whole, den_str)
        else:
            # else return "2 et 3 quart", for example
            strNumber = '{} et {} {}'.format(whole, num, den_str)
        if num > 1 and den != 3:
            # if the numerator is greater than 1 and the denominator
            # is not 3 ("tiers"), add an s for plural
            strNumber += 's'

    return strNumber


def pronounce_number_fr(number, places=2):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'cinq virgule deux'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(number) >= 100:
        # TODO: Support for numbers over 100
        return str(number)

    result = ""
    if number < 0:
        result = "moins "
    number = abs(number)

    if number > 16:
        tens = int(number - int(number) % 10)
        ones = int(number - tens)
        if ones != 0:
            if tens > 10 and tens <= 60 and int(number - tens) == 1:
                result += _NUM_STRING_FR[tens] + "-et-" + _NUM_STRING_FR[ones]
            elif number == 71:
                result += "soixante-et-onze"
            elif tens == 70:
                result += _NUM_STRING_FR[60] + "-"
                if ones < 7:
                    result += _NUM_STRING_FR[10 + ones]
                else:
                    result += _NUM_STRING_FR[10] + "-" + _NUM_STRING_FR[ones]
            elif tens == 90:
                result += _NUM_STRING_FR[80] + "-"
                if ones < 7:
                    result += _NUM_STRING_FR[10 + ones]
                else:
                    result += _NUM_STRING_FR[10] + "-" + _NUM_STRING_FR[ones]
            else:
                result += _NUM_STRING_FR[tens] + "-" + _NUM_STRING_FR[ones]
        else:
            if number == 80:
                result += "quatre-vingts"
            else:
                result += _NUM_STRING_FR[tens]
    else:
        result += _NUM_STRING_FR[int(number)]

    # Deal with decimal part
    if not number == int(number) and places > 0:
        if abs(number) < 1.0 and (result == "moins " or not result):
            result += "zéro"
        result += " virgule"
        _num_str = str(number)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + _NUM_STRING_FR[int(char)]
    return result


def _number_parse_fr(words, i):
    """ Parses a list of words to find a number
    Takes in a list of words (strings without whitespace) and
    extracts a number that starts at the given index.
    Args:
        words (array): the list to extract a number from
        i (int): the index in words where to look for the number
    Returns:
        tuple with number, index of next word after the number.

        Returns None if no number was found.
    """

    def cte_fr(i, s):
        # Check if string s is equal to words[i].
        # If it is return tuple with s, index of next word.
        # If it is not return None.
        if i < len(words) and s == words[i]:
            return s, i + 1
        return None

    def number_word_fr(i, mi, ma):
        # Check if words[i] is a number in _NUMBERS_FR between mi and ma.
        # If it is return tuple with number, index of next word.
        # If it is not return None.
        if i < len(words):
            val = _NUMBERS_FR.get(words[i])
            # Numbers [1-16,20,30,40,50,60,70,80,90,100,1000]
            if val is not None:
                if val >= mi and val <= ma:
                    return val, i + 1
                else:
                    return None
            # The number may be hyphenated (numbers [17-999])
            splitWord = words[i].split('-')
            if len(splitWord) > 1:
                val1 = _NUMBERS_FR.get(splitWord[0])
                if val1:
                    i1 = 0
                    val2 = 0
                    val3 = 0
                    if val1 < 10 and splitWord[1] == "cents":
                        val1 = val1 * 100
                        i1 = 2

                    # For [81-99], e.g. "quatre-vingt-deux"
                    if len(splitWord) > i1 and splitWord[0] == "quatre" and \
                            splitWord[1] == "vingt":
                        val1 = 80
                        i1 += 2

                    # We still found a number
                    if i1 == 0:
                        i1 = 1

                    if len(splitWord) > i1:
                        # For [21,31,41,51,61,71]
                        if len(splitWord) > i1 + 1 and splitWord[i1] == "et":
                            val2 = _NUMBERS_FR.get(splitWord[i1 + 1])
                            if val2 is not None:
                                i1 += 2
                        # For [77-79],[97-99] e.g. "soixante-dix-sept"
                        elif splitWord[i1] == "dix" and \
                                len(splitWord) > i1 + 1:
                            val2 = _NUMBERS_FR.get(splitWord[i1 + 1])
                            if val2 is not None:
                                val2 += 10
                                i1 += 2
                        else:
                            val2 = _NUMBERS_FR.get(splitWord[i1])
                            if val2 is not None:
                                i1 += 1
                                if len(splitWord) > i1:
                                    val3 = _NUMBERS_FR.get(splitWord[i1])
                                    if val3 is not None:
                                        i1 += 1

                        if val2:
                            if val3:
                                val = val1 + val2 + val3
                            else:
                                val = val1 + val2
                        else:
                            return None
                    if i1 == len(splitWord) and val and ma >= val >= mi:
                        return val, i + 1

        return None

    def number_1_99_fr(i):
        # Check if words[i] is a number between 1 and 99.
        # If it is return tuple with number, index of next word.
        # If it is not return None.

        # Is it a number between 1 and 16?
        result1 = number_word_fr(i, 1, 16)
        if result1:
            return result1

        # Is it a number between 10 and 99?
        result1 = number_word_fr(i, 10, 99)
        if result1:
            val1, i1 = result1
            result2 = cte_fr(i1, "et")
            # If the number is not hyphenated [21,31,41,51,61,71]
            if result2:
                i2 = result2[1]
                result3 = number_word_fr(i2, 1, 11)
                if result3:
                    val3, i3 = result3
                    return val1 + val3, i3
            return result1

        # It is not a number
        return None

    def number_1_999_fr(i):
        # Check if words[i] is a number between 1 and 999.
        # If it is return tuple with number, index of next word.
        # If it is not return None.

        # Is it 100 ?
        result = number_word_fr(i, 100, 100)

        # Is it [200,300,400,500,600,700,800,900]?
        if not result:
            resultH1 = number_word_fr(i, 2, 9)
            if resultH1:
                valH1, iH1 = resultH1
                resultH2 = number_word_fr(iH1, 100, 100)
                if resultH2:
                    iH2 = resultH2[1]
                    result = valH1 * 100, iH2

        if result:
            val1, i1 = result
            result2 = number_1_99_fr(i1)
            if result2:
                val2, i2 = result2
                return val1 + val2, i2
            else:
                return result

        # Is it hyphenated? [101-999]
        result = number_word_fr(i, 101, 999)
        if result:
            return result

        # [1-99]
        result = number_1_99_fr(i)
        if result:
            return result

        return None

    def number_1_999999_fr(i):
        """ Find a number in a list of words
        Checks if words[i] is a number between 1 and 999,999.

        Args:
            i (int): the index in words where to look for the number
        Returns:
            tuple with number, index of next word after the number.

            Returns None if no number was found.
        """

        # check for zero
        result1 = number_word_fr(i, 0, 0)
        if result1:
            return result1

        # check for [1-999]
        result1 = number_1_999_fr(i)
        if result1:
            val1, i1 = result1
        else:
            val1 = 1
            i1 = i
        # check for 1000
        result2 = number_word_fr(i1, 1000, 1000)
        if result2:
            # it's [1000-999000]
            i2 = result2[1]
            # check again for [1-999]
            result3 = number_1_999_fr(i2)
            if result3:
                val3, i3 = result3
                return val1 * 1000 + val3, i3
            else:
                return val1 * 1000, i2
        elif result1:
            return result1
        return None

    return number_1_999999_fr(i)


def _get_ordinal_fr(word):
    """ Get the ordinal number
    Takes in a word (string without whitespace) and
    extracts the ordinal number.
    Args:
        word (string): the word to extract the number from
    Returns:
        number (int)

        Returns None if no ordinal number was found.
    """
    if word:
        for ordinal in _ORDINAL_ENDINGS_FR:
            if word[0].isdigit() and ordinal in word:
                result = word.replace(ordinal, "")
                if result.isdigit():
                    return int(result)

    return None


def _number_ordinal_fr(words, i):
    """ Find an ordinal number in a list of words
    Takes in a list of words (strings without whitespace) and
    extracts an ordinal number that starts at the given index.
    Args:
        words (array): the list to extract a number from
        i (int): the index in words where to look for the ordinal number
    Returns:
        tuple with ordinal number (str),
        index of next word after the number (int).

        Returns None if no ordinal number was found.
    """
    val1 = None
    strOrd = ""
    # it's already a digit, normalize to "1er" or "5e"
    val1 = _get_ordinal_fr(words[i])
    if val1 is not None:
        if val1 == 1:
            strOrd = "1er"
        else:
            strOrd = str(val1) + "e"
        return strOrd, i + 1

    # if it's a big number the beginning should be detected as a number
    result = _number_parse_fr(words, i)
    if result:
        val1, i = result
    else:
        val1 = 0

    if i < len(words):
        word = words[i]
        if word in ["premier", "première"]:
            strOrd = "1er"
        elif word == "second":
            strOrd = "2e"
        elif word.endswith("ième"):
            val2 = None
            word = word[:-4]
            # centième
            if word == "cent":
                if val1:
                    strOrd = str(val1 * 100) + "e"
                else:
                    strOrd = "100e"
            # millième
            elif word == "mill":
                if val1:
                    strOrd = str(val1 * 1000) + "e"
                else:
                    strOrd = "1000e"
            else:
                # "cinquième", "trente-cinquième"
                if word.endswith("cinqu"):
                    word = word[:-1]
                # "neuvième", "dix-neuvième"
                elif word.endswith("neuv"):
                    word = word[:-1] + "f"
                result = _number_parse_fr([word], 0)
                if not result:
                    # "trentième", "douzième"
                    word = word + "e"
                    result = _number_parse_fr([word], 0)
                if result:
                    val2, i = result
                if val2 is not None:
                    strOrd = str(val1 + val2) + "e"
        if strOrd:
            return strOrd, i + 1

    return None


def extract_number_fr(text, short_scale=True, ordinals=False):
    """Takes in a string and extracts a number.
    Args:
        text (str): the string to extract a number from
    Returns:
        (str): The number extracted or the original text.
    """
    # TODO: short_scale and ordinals don't do anything here.
    # The parameters are present in the function signature for API compatibility
    # reasons.
    # normalize text, keep articles for ordinals versus fractionals
    text = normalize_fr(text, False)
    # split words by whitespace
    aWords = text.split()
    count = 0
    result = None
    add = False
    while count < len(aWords):
        val = None
        word = aWords[count]
        wordNext = ""
        wordPrev = ""
        if count < (len(aWords) - 1):
            wordNext = aWords[count + 1]
        if count > 0:
            wordPrev = aWords[count - 1]

        if word in _ARTICLES_FR:
            count += 1
            continue
        if word in ["et", "plus", "+"]:
            count += 1
            add = True
            continue

        # is current word a numeric number?
        if word.isdigit():
            val = int(word)
            count += 1
        elif is_numeric(word):
            val = float(word)
            count += 1
        elif wordPrev in _ARTICLES_FR and _get_ordinal_fr(word):
            val = _get_ordinal_fr(word)
            count += 1
        # is current word the denominator of a fraction?
        elif is_fractional_fr(word):
            val = is_fractional_fr(word)
            count += 1

        # is current word the numerator of a fraction?
        if val and wordNext:
            valNext = is_fractional_fr(wordNext)
            if valNext:
                val = float(val) * valNext
                count += 1

        if not val:
            count += 1
            # is current word a numeric fraction like "2/3"?
            aPieces = word.split('/')
            # if (len(aPieces) == 2 and is_numeric(aPieces[0])
            #   and is_numeric(aPieces[1])):
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        # is current word followed by a decimal value?
        if wordNext == "virgule":
            zeros = 0
            newWords = aWords[count + 1:]
            # count the number of zeros after the decimal sign
            for word in newWords:
                if word == "zéro" or word == "0":
                    zeros += 1
                else:
                    break
            afterDotVal = None
            # extract the number after the zeros
            if newWords[zeros].isdigit():
                afterDotVal = newWords[zeros]
                countDot = count + zeros + 2
            # if a number was extracted (since comma is also a
            # punctuation sign)
            if afterDotVal:
                count = countDot
                if not val:
                    val = 0
                # add the zeros
                afterDotString = zeros * "0" + afterDotVal
                val = float(str(val) + "." + afterDotString)
        if val:
            if add:
                result += val
                add = False
            else:
                result = val

    return result or False


def is_fractional_fr(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.
    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction
    """
    input_str = input_str.lower()

    if input_str != "tiers" and input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "quarts"

    aFrac = ["entier", "demi", "tiers", "quart", "cinquième", "sixième",
             "septième", "huitième", "neuvième", "dixième", "onzième",
             "douzième", "treizième", "quatorzième", "quinzième", "seizième",
             "dix-septième", "dix-huitième", "dix-neuvième", "vingtième"]

    if input_str in aFrac:
        return 1.0 / (aFrac.index(input_str) + 1)
    if _get_ordinal_fr(input_str):
        return 1.0 / _get_ordinal_fr(input_str)
    if input_str == "trentième":
        return 1.0 / 30
    if input_str == "centième":
        return 1.0 / 100
    if input_str == "millième":
        return 1.0 / 1000

    return False


def normalize_fr(text, remove_articles=True):
    """ French string normalization """
    text = text.lower()
    words = text.split()  # this also removed extra spaces
    normalized = ""
    i = 0
    while i < len(words):
        # remove articles
        if remove_articles and words[i] in _ARTICLES_FR:
            i += 1
            continue
        if remove_articles and words[i][:2] in ["l'", "d'"]:
            words[i] = words[i][2:]
        # remove useless punctuation signs
        if words[i] in ["?", "!", ";", "…"]:
            i += 1
            continue
        # Normalize ordinal numbers
        if i > 0 and words[i - 1] in _ARTICLES_FR:
            result = _number_ordinal_fr(words, i)
            if result is not None:
                val, i = result
                normalized += " " + str(val)
                continue
        # Convert numbers into digits
        result = _number_parse_fr(words, i)
        if result is not None:
            val, i = result
            normalized += " " + str(val)
            continue

        normalized += " " + words[i]
        i += 1

    return normalized[1:]  # strip the initial space



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from collections import OrderedDict

from ovos_number_parser.util import (convert_to_mixed_fraction, look_for_fractions, invert_dict,
                                     is_numeric, tokenize, partition_list, Token, ReplaceableNumber)

_NUM_STRING_UK = {
    0: "нуль",
    1: "один",
    2: "два",
    3: "три",
    4: "чотири",
    5: "п'ять",
    6: "шість",
    7: "сім",
    8: "вісім",
    9: "дев'ять",
    10: "десять",
    11: "одинадцять",
    12: "дванадцять",
    13: "тринадцять",
    14: "чотирнадцять",
    15: "п'ятнадцять",
    16: "шістнадцять",
    17: "сімнадцять",
    18: "вісімнадцять",
    19: "дев'ятнадцять",
    20: "двадцять",
    30: "тридцять",
    40: "сорок",
    50: "п'ятдесят",
    60: "шістдесят",
    70: "сімдесят",
    80: "вісімдесят",
    90: "дев'яносто",
    100: "сто",
    200: "двісті",
    300: "триста",
    400: "чотириста",
    500: "п'ятсот",
    600: "шістсот",
    700: "сімсот",
    800: "вісімсот",
    900: "дев'ятсот"
}

_PLURALS = {
    'двох': 2, 'двум': 2, 'двома': 2, 'дві': 2, "двоє": 2, "двійка": 2,
    'обидва': 2, 'обидвох': 2, 'обидві': 2, 'обох': 2, 'обома': 2, 'обом': 2,
    'пара': 2, 'пари': 2, 'парою': 2, 'парами': 2, 'парі': 2, 'парах': 2, 'пару': 2,
    'трьох': 3, 'трьома': 3, 'трьом': 3,
    'чотирьох': 4, 'чотирьом': 4, 'чотирма': 4,
    "п'ятьох": 5, "п'ятьом": 5, "п'ятьома": 5,
    "шістьом": 6, "шести": 6, "шістьох": 6, "шістьма": 6, "шістьома": 6,
    "семи": 7, "сімом": 7, "сімох": 7, "сімома": 7, "сьома": 7,
    "восьми": 8, "вісьмох": 8, "вісьмом": 8, "вісьма": 8, "вісьмома": 8,
    "дев'яти": 9, "дев'ятьох": 9, "дев'ятьом": 9, "дев'ятьма": 9,
    "десяти": 10, "десятьох": 10, "десятьма": 10, "десятьома": 10,
    "сорока": 40,
    "сот": 100, "сотень": 100, "сотні": 100, "сотня": 100,
    "двохсот": 200, "двомстам": 200, "двомастами": 200, "двохстах": 200,
    "тисяч": 1000, "тисячі": 1000, "тисячу": 1000, "тисячах": 1000,
    "тисячами": 1000, "тисячею": 1000
}

_FRACTION_STRING_UK = {
    2: "друга",
    3: "третя",
    4: "четверта",
    5: "п'ята",
    6: "шоста",
    7: "сьома",
    8: "восьма",
    9: "дев'ята",
    10: "десята",
    11: "одинадцята",
    12: "дванадцята",
    13: "тринадцята",
    14: "чотирнадцята",
    15: "п'ятнадцята",
    16: "шістнадцята",
    17: "сімнадцята",
    18: "вісімнадцята",
    19: "дев'ятнадцята",
    20: "двадцята",
    30: "тридцята",
    40: "сорокова",
    50: "п'ятдесята",
    60: "шістдесята",
    70: "сімдесята",
    80: "вісімдесята",
    90: "дев'яноста",
    1e2: "сота",
    1e3: "тисячна",
    1e6: "мільйонна",
    1e9: "мільярдна",
    1e-12: "більйонна",
}

_SHORT_SCALE_UK = OrderedDict([
    (1e3, "тисяча"),
    (1e6, "мільйон"),
    (1e9, "мільярд"),
    (1e18, "трильйон"),
    (1e12, "більйон"),
    (1e15, "квадрилліон"),
    (1e18, "квінтиліон"),
    (1e21, "секстильйон"),
    (1e24, "септилліон"),
    (1e27, "октиліон"),
    (1e30, "нонільйон"),
    (1e33, "дециліон"),
    (1e36, "ундеціліон"),
    (1e39, "дуодециліон"),
    (1e42, "тредециліон"),
    (1e45, "кваттордециліон"),
    (1e48, "квіндециліон"),
    (1e51, "сексдециліон"),
    (1e54, "септендециліон"),
    (1e57, "октодециліон"),
    (1e60, "новемдециліон"),
    (1e63, "вігінтильйон"),
    (1e66, "унвігінтільйон"),
    (1e69, "дуовігінтильйон"),
    (1e72, "тревігінтильйон"),
    (1e75, "кватторвігінтільйон"),
    (1e78, "квінвігінтильйон"),
    (1e81, "секснвігінтіліон"),
    (1e84, "септенвігінтильйон"),
    (1e87, "октовігінтиліон"),
    (1e90, "новемвігінтільйон"),
    (1e93, "тригінтильйон"),
])

_LONG_SCALE_UK = OrderedDict([
    (1e3, "тисяча"),
    (1e6, "мільйон"),
    (1e9, "мільярд"),
    (1e12, "більйон"),
    (1e15, "біліард"),
    (1e18, "трильйон"),
    (1e21, "трильярд"),
    (1e24, "квадрилліон"),
    (1e27, "квадрільярд"),
    (1e30, "квінтиліон"),
    (1e33, "квінтільярд"),
    (1e36, "секстильйон"),
    (1e39, "секстильярд"),
    (1e42, "септилліон"),
    (1e45, "септільярд"),
    (1e48, "октиліон"),
    (1e51, "октільярд"),
    (1e54, "нонільйон"),
    (1e57, "нонільярд"),
    (1e60, "дециліон"),
    (1e63, "дециліард"),
    (1e66, "ундеціліон"),
    (1e72, "дуодециліон"),
    (1e78, "тредециліон"),
    (1e84, "кваттордециліон"),
    (1e90, "квіндециліон"),
    (1e96, "сексдециліон"),
    (1e102, "септендециліон"),
    (1e108, "октодециліон"),
    (1e114, "новемдециліон"),
    (1e120, "вігінтильйон"),
])

_ORDINAL_BASE_UK = {
    1: "перший",
    2: "другий",
    3: "третій",
    4: "четвертий",
    5: "п'ятий",
    6: "шостий",
    7: "сьомий",
    8: "восьмий",
    9: "дев'ятий",
    10: "десятий",
    11: "одинадцятий",
    12: "дванадцятий",
    13: "тринадцятий",
    14: "чотирнадцятий",
    15: "п'ятнадцятий",
    16: "шістнадцятий",
    17: "сімнадцятий",
    18: "вісімнадцятий",
    19: "дев'ятнадцятий",
    20: "двадцятий",
    30: "тридцятий",
    40: "сороковий",
    50: "п'ятдесятий",
    60: "шістдесятий",
    70: "сімдесятий",
    80: "вісімдесятий",
    90: "дев'яностий",
    1e2: "сотий",
    2e2: "двохсотий",
    3e2: "трьохсотий",
    4e2: "чотирисотий",
    5e2: "п'ятисотий",
    6e2: "шістсотий",
    7e2: "семисотий",
    8e2: "восьмисотий",
    9e2: "дев'ятисотий",
    1e3: "тисячний"
}

_SHORT_ORDINAL_UK = {
    1e6: "мільйон",
    1e9: "мільярд",
    1e18: "трильйон",
    1e15: "квадрилліон",
    1e18: "квінтильйон",
    1e21: "секстильйон",
    1e24: "септилліон",
    1e27: "октиліон",
    1e30: "нонільйон",
    1e33: "дециліон",
    1e36: "ундеціліон",
    1e39: "дуодециліон",
    1e42: "тредециліон",
    1e45: "кваттордециліон",
    1e48: "квіндециліон",
    1e51: "сексдециліон",
    1e54: "септендециліон",
    1e57: "октодециліон",
    1e60: "новемдециліон",
    1e63: "вігінтильйон"
}
_SHORT_ORDINAL_UK.update(_ORDINAL_BASE_UK)

_LONG_ORDINAL_UK = {
    1e6: "мільйон",
    1e9: "мільярд",
    1e12: "більйон",
    1e15: "біліард",
    1e18: "трильйон",
    1e21: "трильярд",
    1e24: "квадрилліон",
    1e27: "квадрильярд",
    1e30: "квінтиліон",
    1e33: "квінтільярд",
    1e36: "секстильйон",
    1e39: "секстильярд",
    1e42: "септилліон",
    1e45: "септільярд",
    1e48: "октиліон",
    1e51: "октільярд",
    1e54: "нонільйон",
    1e57: "нонільярд",
    1e60: "дециліон",
    1e63: "дециліард",
    1e66: "ундеціліон",
    1e72: "дуодециліон",
    1e78: "тредециліон",
    1e84: "кваттордециліон",
    1e90: "квіндециліон",
    1e96: "сексдециліон",
    1e102: "септендециліон",
    1e108: "октодециліон",
    1e114: "новемдециліон",
    1e120: "вігінтильйон"
}
_LONG_ORDINAL_UK.update(_ORDINAL_BASE_UK)


def generate_plurals_uk(originals):
    """
    Return a new set or dict containing the plural form of the original values,
    Generate different cases of values

    In English this means all with 's' appended to them.

    Args:
        originals set(str) or dict(str, any): values to pluralize

    Returns:
        set(str) or dict(str, any)

    """
    suffixes = ["а", "ах", "их", "ам", "ами", "ів",
                "ям", "ох", "и", "на", "ни", "і", "ні",
                "ий", "ний", 'ьох', 'ьома', 'ьом', 'ох',
                'ум', 'ма', 'ом']
    if isinstance(originals, dict):
        thousand = {"тисяч": 1000, "тисячі": 1000, "тисячу": 1000, "тисячах": 1000}
        hundred = {"сотня": 100, "сотні": 100, "сотень": 100}
        result_dict = {key + suffix: value for key, value in originals.items() for suffix in suffixes}
        result_dict.update(thousand)
        result_dict.update(hundred)
        return result_dict
    thousand = ["тисяч", "тисячі", "тисячу", "тисячах"]
    result_dict = {value + suffix for value in originals for suffix in suffixes}
    result_dict.update(thousand)
    return {value + suffix for value in originals for suffix in suffixes}


# negate next number (-2 = 0 - 2)
_NEGATIVES = {"мінус"}

# sum the next number (twenty two = 20 + 2)
_SUMS = {"двадцять", "20", "тридцять", "30", "сорок", "40", "п'ятдесят", "50",
         "шістдесят", "60", "сімдесят", "70", "вісімдесят", "80", "дев'яносто", "90",
         "сто", "100", "двісті", "200", "триста", "300", "чотириста", "400",
         "п'ятсот", "500", "шістсот", "600", "сімсот", "700", "вісімсот", "800",
         "дев'ятсот", "900"}

_MULTIPLIES_LONG_SCALE_UK = set(_LONG_SCALE_UK.values()) | \
                            generate_plurals_uk(_LONG_SCALE_UK.values())

_MULTIPLIES_SHORT_SCALE_UK = set(_SHORT_SCALE_UK.values()) | \
                             generate_plurals_uk(_SHORT_SCALE_UK.values())

# split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
_FRACTION_MARKER = {"і", "та", "з", " "}

# decimal marker ( 1 point 5 = 1 + 0.5)
_DECIMAL_MARKER = {"ціла", "цілих", "точка", "крапка", "кома"}

_STRING_NUM_UK = invert_dict(_NUM_STRING_UK)

_STRING_NUM_UK.update(generate_plurals_uk(_STRING_NUM_UK))
_STRING_NUM_UK.update(_PLURALS)
_STRING_NUM_UK.update({
    "трильйон": 1e18,
    "половина": 0.5, "половиною": 0.5, "половини": 0.5, "половин": 0.5, "половинами": 0.5, "пів": 0.5,
    "одна": 1, "одної": 1, "одній": 1, "одну": 1
})

_WORDS_NEXT_UK = [
    "майбутня", "майбутнє", "майбутній", "майбутньому", "майбутнім", "майбутньої", "майбутнього",
    "нова", "нове", "новий", "нового", "нової", "новим", "новою", "через",
    "наступна", "наступне", "наступний", "наступній", "наступному", "наступним", "наступною",
]
_WORDS_PREV_UK = [
    "попередня", "попередній", "попереднім", "попередньої",
    "попередню", "попереднього", "попередне", "тому",
    "минула", "минулий", "минуле", "минулу", "минулого", "минулій", "минулому",
    "минулої", "минулою", "минулим",
    "та", "той", "ті", "те", "того",
]
_WORDS_CURRENT_UK = [
    "теперішній", "теперішня", "теперішні", "теперішній", "теперішньому",
    "теперішньою", "теперішнім", "теперішнього", "теперішньої",
    "дана", "даний", "дане", "даним", "даною", "даного", "даної", "даному", "даній",
    "поточний", "поточна", "поточні", "поточне", "поточного", "поточної",
    "поточному", "поточній", "поточним", "поточною",
    "нинішній", "нинішня", "нинішнє", "нинішньому", "нинішній",
    "нинішнього", "нинішньої", "нинішнім", "нинішньою",
    "цей", "ця", "це", "цим", "цією", "цьому", "цій"
]
_WORDS_NOW_UK = [
    "тепер",
    "зараз",
]
_WORDS_MORNING_UK = ["ранок", "зранку", "вранці", "ранку"]
_WORDS_DAY_UK = ["вдень", "опівдні"]
_WORDS_EVENING_UK = ["вечер", "ввечері", "увечері", "вечором"]
_WORDS_NIGHT_UK = ["ніч", "вночі"]

_STRING_SHORT_ORDINAL_UK = invert_dict(_SHORT_ORDINAL_UK)
_STRING_LONG_ORDINAL_UK = invert_dict(_LONG_ORDINAL_UK)


def nice_number_uk(number, speech=True, denominators=range(1, 21)):
    """ Ukrainian helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 and a half" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_UK[den]

    if whole == 0:
        return_string = '{} {}'.format(num, den_str)
    elif num == 1 and den == 2:
        return_string = '{} з половиною'.format(whole)
    else:
        return_string = '{} і {} {}'.format(whole, num, den_str)
    if 2 <= den <= 4:
        if 2 <= num <= 4:
            return_string = return_string[:-1] + 'і'
        elif num > 4:
            return_string = return_string[:-1] + 'ій'
    elif den >= 5:
        if 2 <= num <= 4:
            return_string = return_string[:-1] + 'і'
        elif num > 4:
            return_string = return_string[:-1] + 'их'

    return return_string


def pronounce_number_uk(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """
    num = number
    # deal with infinity
    if num == float("inf"):
        return "нескінченність"
    elif num == float("-inf"):
        return "мінус нескінченність"
    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            if ordinals:
                # This handles negative powers separately from the normal
                # handling since each call disables the scientific flag
                if float(n) < 0:
                    first_part = 'мінус ' + pronounce_number_uk(
                        abs(float(n)), places, short_scale, False, ordinals=True)
                else:
                    first_part = pronounce_number_uk(
                        abs(float(n)), places, short_scale, False, ordinals=True)

                if power < 0:
                    second_part = 'мінус ' + pronounce_number_uk(
                        abs(power), places, short_scale, False, ordinals=True)
                else:
                    second_part = pronounce_number_uk(
                        abs(power), places, short_scale, False, ordinals=True)
                if second_part.endswith('ий'):
                    second_part = second_part[:-2] + 'ому'

                return '{} на десять у {} ступені'.format(
                    first_part, second_part)
            else:
                # This handles negative powers separately from the normal
                # handling since each call disables the scientific flag
                return '{}{} на десять у ступені {}{}'.format(
                    'мінус ' if float(n) < 0 else '',
                    pronounce_number_uk(
                        abs(float(n)), places, short_scale, False, ordinals=False),
                    'мінус ' if power < 0 else '',
                    pronounce_number_uk(abs(power), places, short_scale, False, ordinals=False))

    if short_scale:
        number_names = _NUM_STRING_UK.copy()
        number_names.update(_SHORT_SCALE_UK)
    else:
        number_names = _NUM_STRING_UK.copy()
        number_names.update(_LONG_SCALE_UK)

    digits = [number_names[n] for n in range(0, 20)]

    tens = [number_names[n] for n in range(10, 100, 10)]

    if short_scale:
        hundreds = [_SHORT_SCALE_UK[n] for n in _SHORT_SCALE_UK.keys()]
    else:
        hundreds = [_LONG_SCALE_UK[n] for n in _LONG_SCALE_UK.keys()]

    # deal with negative numbers
    result = ""
    if num < 0:
        result = "мінус "
    num = abs(num)

    # check for a direct match
    if num in number_names and not ordinals:
        result += number_names[num]
    else:
        def _sub_thousand(n, ordinals=False):
            assert 0 <= n <= 999
            if n in _SHORT_ORDINAL_UK and ordinals:
                return _SHORT_ORDINAL_UK[n]
            if n <= 19:
                return digits[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                return tens[q - 1] + (" " + _sub_thousand(r, ordinals) if r
                                      else "")
            else:
                q, r = divmod(n, 100)
                return _NUM_STRING_UK[q * 100] + (" " + _sub_thousand(r, ordinals) if r else "")

        def _short_scale(n):
            if n > max(_SHORT_SCALE_UK.keys()):
                return "нескінченність"
            ordi = ordinals

            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000)):
                if not z:
                    continue
                number = _sub_thousand(z, not i and ordi)
                if i:
                    if i >= len(hundreds):
                        return ""
                    if ordi:
                        if i * 1000 in _SHORT_ORDINAL_UK:
                            if z == 1:
                                number = _SHORT_ORDINAL_UK[i * 1000]
                            else:
                                if z > 5:
                                    number = number[:-1] + "и"
                                number += _SHORT_ORDINAL_UK[i * 1000]
                        else:
                            if n not in _SHORT_SCALE_UK:
                                num = int("1" + "0" * (len(str(n)) // 3 * 3))

                                if number[-3:] == "два":
                                    number = number[:-1] + "ох"
                                elif number[-2:] == "ри" or number[-2:] == "ре":
                                    number = number[:-1] + "ьох"
                                elif number[-1:] == "ь":
                                    number = number[:-1] + "и"

                                if _SHORT_SCALE_UK[num].endswith('н'):
                                    number += _SHORT_SCALE_UK[num] + "ний"
                                else:
                                    number += _SHORT_SCALE_UK[num] + "ий"
                            else:
                                if _SHORT_SCALE_UK[n].endswith('н'):
                                    number = _SHORT_SCALE_UK[n] + "ний"
                                else:
                                    number = _SHORT_SCALE_UK[n] + "ий"
                    elif z == 1:
                        number = hundreds[i - 1]
                    else:
                        if i == 1:
                            if z % 10 == 1 and z % 100 // 10 != 1:
                                number = number[:-2] + "на"
                            elif z % 10 == 2 and z % 100 // 10 != 1:
                                number = number[:-1] + "і"
                            number += " " + plural_uk(z, "тисяча", "тисячі", "тисяч")
                        elif 1 <= z % 10 <= 4 and z % 100 // 10 != 1:
                            number += " " + hundreds[i - 1] + "а"
                        else:
                            number += " " + hundreds[i - 1] + "ів"

                res.append(number)
                ordi = False

            return " ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        def _long_scale(n):
            if n >= max(_LONG_SCALE_UK.keys()):
                return "нескінченність"
            ordi = ordinals
            if int(n) != n:
                ordi = False
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000000)):
                if not z:
                    continue
                number = pronounce_number_uk(z, places, True, scientific,
                                             ordinals=ordi and not i)
                # strip off the comma after the thousand
                if i:
                    if i >= len(hundreds):
                        return ""
                    # plus one as we skip 'thousand'
                    # (and 'hundred', but this is excluded by index value)
                    number = number.replace(',', '')

                    if ordi:
                        if (i + 1) * 1000000 in _LONG_ORDINAL_UK:
                            if z == 1:
                                number = _LONG_ORDINAL_UK[
                                    (i + 1) * 1000000]
                            else:
                                number += _LONG_ORDINAL_UK[
                                    (i + 1) * 1000000]
                        else:
                            if n not in _LONG_SCALE_UK:
                                num = int("1" + "0" * (len(str(n)) // 3 * 3))

                                if number[-3:] == "два":
                                    number = number[:-1] + "ох"
                                elif number[-2:] == "ри" or number[-2:] == "ре":
                                    number = number[:-1] + "ьох"
                                elif number[-1:] == "ь":
                                    number = number[:-1] + "и"

                                number += _LONG_SCALE_UK[num] + "ний"
                            else:
                                number = " " + _LONG_SCALE_UK[n] + "ний"
                    elif z == 1:
                        number = hundreds[i]
                    elif z <= 4:
                        number += " " + hundreds[i] + "а"
                    else:
                        number += " " + hundreds[i] + "ів"

                res.append(number)
            return " ".join(reversed(res))

        if short_scale:
            result += _short_scale(num)
        else:
            result += _long_scale(num)

    # deal with scientific notation unpronounceable as number
    if not result and "e" in str(num):
        return pronounce_number_uk(num, places, short_scale, scientific=True)
    # Deal with fractional part
    elif not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "мінус " or not result):
            result += "нуль"
        result += " крапка"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result


def numbers_to_digits_uk(text, short_scale=True, ordinals=False):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numbers should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)

    Returns:
        str
        The original text, with numbers subbed in where appropriate.

    """
    text = text.lower()

    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_uk(tokens, short_scale, ordinals)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)
    return ' '.join(results)


def _extract_numbers_with_text_uk(tokens, short_scale=True,
                                  ordinals=False, fractional_numbers=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_uk(tokens, short_scale,
                                         ordinals, fractional_numbers)

        if not to_replace:
            break
        results.append(to_replace)
        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_uk(tokens, short_scale=True,
                                 ordinals=False, fractional_numbers=True):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = \
        _extract_number_with_text_uk_helper(tokens, short_scale,
                                            ordinals, fractional_numbers)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_uk_helper(tokens,
                                        short_scale=True, ordinals=False,
                                        fractional_numbers=True):
    """
    Helper for _extract_number_with_text_uk.

    This contains the real logic for parsing, but produces
    a result that needs a little cleaning (specific, it may
    contain leading articles that can be trimmed off).

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:

    Returns:
        int or float, [Tokens]

    """
    if fractional_numbers:
        fraction, fraction_text = \
            _extract_fraction_with_text_uk(tokens, short_scale, ordinals)
        if fraction:
            return fraction, fraction_text

        decimal, decimal_text = \
            _extract_decimal_with_text_uk(tokens, short_scale, ordinals)
        if decimal:
            return decimal, decimal_text
    # special_number = [word for word in tokens if word ]
    # short_scale == False
    return _extract_whole_number_with_text_uk(tokens, short_scale, ordinals)


def _extract_fraction_with_text_uk(tokens, short_scale, ordinals):
    """
    Extract fraction numbers from a string.

    This function handles text such as '2 and 3/4'. Note that "one half" or
    similar will be parsed by the whole number function.

    Args:
        tokens [Token]: words and their indexes in the original string.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (int or float, [Token])
        The value found, and the list of relevant tokens.
        (None, None) if no fraction value is found.

    """
    for c in _FRACTION_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_uk(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_uk(partitions[2], short_scale,
                                              ordinals, fractional_numbers=True)

            if not numbers1 or not numbers2:
                return None, None

            # ensure first is not a fraction and second is a fraction
            num1 = numbers1[-1]
            num2 = numbers2[0]
            if num1.value >= 1 and 0 < num2.value < 1:
                return num1.value + num2.value, \
                       num1.tokens + partitions[1] + num2.tokens

    return None, None


def _extract_decimal_with_text_uk(tokens, short_scale, ordinals):
    """
    Extract decimal numbers from a string.

    This function handles text such as '2 point 5'.

    Notes:
        While this is a helper for extract_number_xx, it also depends on
        extract_number_xx, to parse out the components of the decimal.

        This does not currently handle things like:
            number dot number number number

    Args:
        tokens [Token]: The text to parse.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (float, [Token])
        The value found and relevant tokens.
        (None, None) if no decimal value is found.

    """
    for c in _DECIMAL_MARKER:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_uk(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_uk(partitions[2], short_scale,
                                              ordinals, fractional_numbers=False)

            if not numbers1 or not numbers2:
                return None, None

            number = numbers1[-1]
            decimal = numbers2[0]

            # TODO handle number dot number number number
            if "." not in str(decimal.text):
                return number.value + float('0.' + str(decimal.value)), \
                       number.tokens + partitions[1] + decimal.tokens
    return None, None


def _extract_whole_number_with_text_uk(tokens, short_scale, ordinals):
    """
    Handle numbers not handled by the decimal or fraction functions. This is
    generally whole numbers. Note that phrases such as "one half" will be
    handled by this function, while "one and a half" are handled by the
    fraction function.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:

    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.

    """
    number_token = [token for token in tokens if token.word.lower() in _MULTIPLIES_LONG_SCALE_UK]
    if number_token:
        short_scale = False
    multiplies, string_num_ordinal, string_num_scale = \
        _initialize_number_data(short_scale)
    number_words = []  # type: [Token]
    val = False
    prev_val = None
    next_val = None
    to_sum = []
    for idx, token in enumerate(tokens):
        current_val = None
        if next_val:
            next_val = None
            continue

        word = token.word
        if word in word in _NEGATIVES:
            number_words.append(token)
            continue

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        prev_word = _text_uk_inflection_normalize(prev_word, 1)
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""
        next_word = _text_uk_inflection_normalize(next_word, 1)

        # In Ukrainian (?) we do not use suffix (1st,2nd,..) but use point instead (1.,2.,..)
        if is_numeric(word[:-1]) and \
                (word.endswith(".")):
            # explicit ordinals, 1st, 2nd, 3rd, 4th.... Nth
            word = word[:-1]

        # Normalize Ukrainian inflection of numbers (один, одна, одно,...)
        if not ordinals:
            if word not in _STRING_NUM_UK:
                word = _text_uk_inflection_normalize(word, 1)

        if word not in string_num_scale and \
                word not in _STRING_NUM_UK and \
                word not in _SUMS and \
                word not in multiplies and \
                not (ordinals and word in string_num_ordinal) and \
                not is_numeric(word) and \
                not is_fractional_uk(word, word, short_scale=short_scale) and \
                not look_for_fractions(word.split('/')):
            words_only = [token.word for token in number_words]
            if number_words and not all([w in _NEGATIVES for w in words_only]):
                break
            else:
                number_words = []
                continue
        elif word not in multiplies \
                and prev_word not in multiplies \
                and prev_word not in _SUMS \
                and not (ordinals and prev_word in string_num_ordinal) \
                and prev_word not in _NEGATIVES:

            number_words = [token]
        elif prev_word in _SUMS and word in _SUMS:
            number_words = [token]
        else:
            number_words.append(token)
        # is this word already a number ?
        if is_numeric(word):
            if word.isdigit():  # doesn't work with decimals
                val = int(word)
            else:
                val = float(word)
            current_val = val

        # is this word the name of a number ?
        if word in _STRING_NUM_UK:
            val = _STRING_NUM_UK.get(word)
            current_val = val
        elif word in string_num_scale:
            val = string_num_scale.get(word)
            current_val = val
        elif ordinals and word in string_num_ordinal:
            val = string_num_ordinal[word]
            current_val = val
        # is the prev word an ordinal number and current word is one?
        # second one, third one
        if ordinals and prev_word in string_num_ordinal and val == 1:
            val = prev_val
        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if (prev_word in _SUMS and val and val < 10) \
                or (prev_word in _SUMS and val and val < 100 and prev_val >= 100) \
                or all([prev_word in multiplies, val < prev_val if prev_val else False]):
            val = prev_val + val

        # is the prev word a number and should we multiply it?
        multiplies.update({"тисячa", "тисячі", "тисячу", "тисячах", "тисячaми", "тисячею", "тисяч"})
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val

        # пара сотень, три пари пива
        if prev_word in ['пара', 'пари', 'парою', 'парами'] and current_val != 1000.0:
            val = val * 2
        if prev_val in _STRING_NUM_UK.values() and current_val == 100:
            val = prev_val * current_val

        # half cup
        if val is False:
            val = is_fractional_uk(word, word, short_scale=short_scale)
            current_val = val

        # 2 fifths
        if not ordinals:
            next_val = is_fractional_uk(next_word, word, short_scale=short_scale)
            if next_val:
                if not val:
                    val = 1
                val = val * next_val
                number_words.append(tokens[idx + 1])
        if word in ['пара', 'пари', 'парою', 'парами']:
            if prev_val:
                val = val * prev_val
            else:
                val = 2
        # is this a negative number?
        if val and prev_word and prev_word in _NEGATIVES:
            val = 0 - val

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            a_pieces = word.split('/')
            if look_for_fractions(a_pieces):
                val = float(a_pieces[0]) / float(a_pieces[1])
        else:
            # checking if word is digit in order not to substitute
            # existing calculated value
            new_word = re.sub(r'\.', '', word)
            if all([
                prev_word in _SUMS,
                word not in _SUMS,
                new_word.isdigit() is False,
                word not in multiplies,
                current_val >= 10
            ]):
                # Backtrack - we've got numbers we can't sum.
                number_words.pop()
                val = prev_val
                break
            prev_val = val
            if word in multiplies and next_word not in multiplies:
                # handle long numbers
                # six hundred sixty six
                # two million five hundred thousand
                #
                # This logic is somewhat complex, and warrants
                # extensive documentation for the next coder's sake.
                #
                # The current word is a power of ten. `current_val` is
                # its integer value. `val` is our working sum
                # (above, when `current_val` is 1 million, `val` is
                # 2 million.)
                #
                # We have a dict `string_num_scale` containing [value, word]
                # pairs for "all" powers of ten: string_num_scale[10] == "ten.
                #
                # We need go over the rest of the tokens, looking for other
                # powers of ten. If we find one, we compare it with the current
                # value, to see if it's smaller than the current power of ten.
                #
                # Numbers which are not powers of ten will be passed over.
                #
                # If all the remaining powers of ten are smaller than our
                # current value, we can set the current value aside for later,
                # and begin extracting another portion of our final result.
                # For example, suppose we have the following string.
                # The current word is "million".`val` is 9000000.
                # `current_val` is 1000000.
                #
                #    "nine **million** nine *hundred* seven **thousand**
                #     six *hundred* fifty seven"
                #
                # Iterating over the rest of the string, the current
                # value is larger than all remaining powers of ten.
                #
                # The if statement passes, and nine million (9000000)
                # is appended to `to_sum`.
                #
                # The main variables are reset, and the main loop begins
                # assembling another number, which will also be appended
                # under the same conditions.
                #
                # By the end of the main loop, to_sum will be a list of each
                # "place" from 100 up: [9000000, 907000, 600]
                #
                # The final three digits will be added to the sum of that list
                # at the end of the main loop, to produce the extracted number:
                #
                #    sum([9000000, 907000, 600]) + 57
                # == 9,000,000 + 907,000 + 600 + 57
                # == 9,907,657
                #
                # >>> foo = "nine million nine hundred seven thousand six
                #            hundred fifty seven"
                # >>> extract_number(foo)
                # 9907657
                time_to_sum = True
                for other_token in tokens[idx + 1:]:
                    if other_token.word in multiplies:
                        if string_num_scale[other_token.word] >= current_val:
                            time_to_sum = False
                        else:
                            continue
                    if not time_to_sum:
                        break
                if time_to_sum:
                    to_sum.append(val)
                    val = 0
                    prev_val = 0

    if val is not None and to_sum:
        val += sum(to_sum)
    return val, number_words


def _initialize_number_data(short_scale):
    """
    Generate dictionaries of words to numbers, based on scale.

    This is a helper function for _extract_whole_number.

    Args:
        short_scale boolean:

    Returns:
        (set(str), dict(str, number), dict(str, number))
        multiplies, string_num_ordinal, string_num_scale

    """
    multiplies = _MULTIPLIES_SHORT_SCALE_UK if short_scale \
        else _MULTIPLIES_LONG_SCALE_UK

    string_num_ordinal_uk = _STRING_SHORT_ORDINAL_UK if short_scale \
        else _STRING_LONG_ORDINAL_UK

    string_num_scale_uk = _SHORT_SCALE_UK if short_scale else _LONG_SCALE_UK
    string_num_scale_uk = invert_dict(string_num_scale_uk)
    string_num_scale_uk.update(generate_plurals_uk(string_num_scale_uk))
    return multiplies, string_num_ordinal_uk, string_num_scale_uk


def extract_number_uk(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """
    return _extract_number_with_text_uk(tokenize(text.lower()),
                                        short_scale, ordinals).value


def is_fractional_uk(input_str, word, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    fractions = {"ціла": 1}
    # endings for creation different cases and plurals in different cases
    ending = ['ої', 'е', 'их', 'ою', 'і', 'ими', 'ій']
    for num in _FRACTION_STRING_UK.keys():  # Numbers from 2 to 1 hundred, more is not usually used in common speech
        if num > 1:
            fractions[str(_FRACTION_STRING_UK[num])] = num
            for end in ending:
                new_fraction_number = _FRACTION_STRING_UK[num][:-1] + end
                fractions[new_fraction_number] = num
    fractions.update({
        "половина": 2, "половиною": 2, "половини": 2, "половин": 2, "половинами": 2, "пів": 2,
        "шоста": 6,
        "третина": 1 / 3, "треть": 1 / 3, "треті": 3, "третьої": 3,
        "чверті": 4, "чверть": 0.25, "чвертю": 0.25
    })
    if input_str.lower() in fractions.keys():
        if word == input_str:
            return fractions[input_str.lower()]
        elif word not in _STRING_NUM_UK:
            return fractions[input_str.lower()]
        else:
            return 1.0 / fractions[input_str.lower()]
    return False


def _text_uk_inflection_normalize(word, arg):
    """
    Ukrainian Inflection normalizer.

    This try to normalize known inflection. This function is called
    from multiple places, each one is defined with arg.

    Args:
        word [Word]
        arg [Int]

    Returns:
        word [Word]

    """

    if arg == 1:  # _extract_whole_number_with_text_uk
        if word in ["одна", "одним", "одно", "одною", "одного", "одної", "одному", "одній", "одного", "одну"]:
            return "один"
        return _plurals_normalizer(word)

    elif arg == 2:  # extract_datetime_uk
        if word in ["година", "години", "годин", "годину", "годин", "годинами"]:
            return "година"
        if word in ["хвилина", "хвилини", "хвилину", "хвилин", "хвилька"]:
            return "хвилина"
        if word in ["секунд", "секунди", "секундами", "секунду", "секунд", "сек"]:
            return "секунда"
        if word in ["днів", "дні", "днями", "дню", "днем", "днями"]:
            return "день"
        if word in ["тижні", "тижнів", "тижнями", "тиждень", "тижня"]:
            return "тиждень"
        if word in ["місяцем", "місяці", "місяця", "місяцях", "місяцем", "місяцями", "місяців"]:
            return "місяць"
        if word in ["року", "роки", "році", "роках", "роком", "роками", "років"]:
            return "рік"
        if word in _WORDS_MORNING_UK:
            return "вранці"
        if word in ["опівдні", "півдня"]:
            return "південь"
        if word in _WORDS_EVENING_UK:
            return "ввечері"
        if word in _WORDS_NIGHT_UK:
            return "ніч"
        if word in ["вікенд", "вихідних", "вихідними"]:
            return "вихідні"
        if word in ["столітті", "століттях", "століть"]:
            return "століття"
        if word in ["десятиліття", "десятиліть", "десятиліттях"]:
            return "десятиліття"
        if word in ["столітті", "століттях", "століть"]:
            return "століття"

        # Week days
        if word in ["понеділка", "понеділки"]:
            return "понеділок"
        if word in ["вівторка", "вівторки"]:
            return "вівторок"
        if word in ["середу", "середи"]:
            return "среда"
        if word in ["четверга"]:
            return "четвер"
        if word in ["п'ятницю", "п'ятниці"]:
            return "п'ятниця"
        if word in ["суботу", "суботи"]:
            return "субота"
        if word in ["неділю", "неділі"]:
            return "неділя"

        # Months
        if word in ["лютому", "лютого", "лютим"]:
            return "лютий"
        if word in ["листопада", "листопаді", "листопадом"]:
            return "листопад"
        tmp = ''
        if word[-3:] in ["ого", "ому"]:
            tmp = word[:-3] + "ень"
        elif word[-2:] in ["ні", "ня"]:
            tmp = word[:-2] + "ень"
    return word


def _plurals_normalizer(word):
    """
    Ukrainian Plurals normalizer.

    This function normalizes plural endings of numerals
    including different case variations.
    Uses _PLURALS dictionary with exceptions that can not
    be covered by rules.
    Args:
        word [Word]

    Returns:
        word [Word]

    """
    if word not in _STRING_NUM_UK:
        # checking for plurals 2-10
        for key, value in _PLURALS.items():
            if word == key:
                return _NUM_STRING_UK[value]

        # checking for plurals 11-19
        case_endings = ['надцяти', 'надцятим', 'надцятими',
                        'надцятьох', 'надцятьма', 'надцятьома', 'надцятьом']
        plural_case = ''.join([case for case in case_endings if case in word])
        if plural_case:
            if 'один' in word:
                return "одинадцять"
            word = word.replace(plural_case, '') + 'надцять'
            return word

        # checking for plurals 20,30
        case_endings = ['дцяти', 'дцятим', 'дцятими',
                        'дцятьох', 'дцятьма', 'дцятьома', 'дцятьом']
        plural_case = ''.join([case for case in case_endings if case in word])
        if plural_case:
            word = word.replace(plural_case, '') + 'дцять'
            return word

        # checking for plurals 50, 60, 70, 80
        case_endings = ['десятьох', 'десяти', 'десятьом',
                        'десятьма', 'десятьома']
        plural_case = ''.join([case for case in case_endings if case in word])
        if plural_case:
            word = word.replace(plural_case, '') + 'десят'
            return word

        # checking for plurals 90, 100
        case_endings = ['стам', 'стами', 'стах',
                        'стами', 'ста', 'сот']
        plural_case = ''.join([case for case in case_endings if case in word])
        if plural_case:
            word = word.replace(plural_case, '')
            for key, value in _PLURALS.items():
                if word == key:
                    firs_part = _NUM_STRING_UK[value]
                    if value in [3, 4]:
                        word = firs_part + 'ста'
                    elif value in [5, 6, 9]:
                        word = firs_part[:-1] + 'сот'
                    elif value in [7, 8]:
                        word = firs_part + 'сот'
                    return word
            return word
    return word


def plural_uk(num: int, one: str, few: str, many: str):
    num %= 100
    if num // 10 == 1:
        return many
    if num % 10 == 1:
        return one
    if 2 <= num % 10 <= 4:
        return few
    return many



from math import floor

from ovos_number_parser.util import (invert_dict, convert_to_mixed_fraction, tokenize,
                                     ReplaceableNumber, Token, look_for_fractions)


_ARTICLES = {'en', 'et'}

_DA_NUMBERS = {
    'nul': 0,
    'en': 1,
    'et': 1,
    'to': 2,
    'tre': 3,
    'fire': 4,
    'fem': 5,
    'seks': 6,
    'syv': 7,
    'otte': 8,
    'ni': 9,
    'ti': 10,
    'elve': 11,
    'tolv': 12,
    'tretten': 13,
    'fjorten': 14,
    'femten': 15,
    'seksten': 16,
    'sytten': 17,
    'atten': 18,
    'nitten': 19,
    'tyve': 20,
    'enogtyve': 21,
    'toogtyve': 22,
    'treogtyve': 23,
    'fireogtyve': 24,
    'femogtyve': 25,
    'seksogtyve': 26,
    'syvogtyve': 27,
    'otteogtyve': 28,
    'niogtyve': 29,
    'tredive': 30,
    'enogtredive': 31,
    'fyrre': 40,
    'halvtres': 50,
    'tres': 60,
    'halvfjers': 70,
    'firs': 80,
    'halvfems': 90,
    'hunderede': 100,
    'ethunderede': 100,
    'tohundrede': 200,
    'trehundrede': 300,
    'firehundrede': 400,
    'femhundrede': 500,
    'sekshundrede': 600,
    'syvhundrede': 700,
    'ottehundrede': 800,
    'nihundrede': 900,
    'tusinde': 1000,
    'ettusinde': 1000,
    'totusinde': 2000,
    'tretusinde': 3000,
    'firetusinde': 4000,
    'femtusinde': 5000,
    'sekstusinde': 6000,
    'syvtusinde': 7000,
    'ottetusinde': 8000,
    'nitusinde': 9000,
    'titusinde': 10000,
    'million': 1000000
}

_MONTHS_DA = ['januar', 'februar', 'marts', 'april', 'maj', 'juni',
              'juli', 'august', 'september', 'oktober', 'november',
              'december']

_NUM_STRING_DA = {
    0: 'nul',
    1: 'en',
    2: 'to',
    3: 'tre',
    4: 'fire',
    5: 'fem',
    6: 'seks',
    7: 'syv',
    8: 'otte',
    9: 'ni',
    10: 'ti',
    11: 'elve',
    12: 'tolv',
    13: 'tretten',
    14: 'fjorten',
    15: 'femten',
    16: 'seksten',
    17: 'sytten',
    18: 'atten',
    19: 'nitten',
    20: 'tyve',
    30: 'tredive',
    40: 'fyrre',
    50: 'halvtres',
    60: 'tres',
    70: 'halvfjers',
    80: 'firs',
    90: 'halvfems',
    100: 'hundrede',
    200: 'tohundrede',
    300: 'trehundrede',
    400: 'firehundrede',
    500: 'femhundrede',
    600: 'sekshundrede',
    700: 'syvhundrede',
    800: 'ottehundrede',
    900: 'nihundrede',
    1000: 'tusinde',
    1000000: 'million'
}

_NUM_POWERS_OF_TEN = [
    'hundred',
    'tusind',
    'million',
    'milliard',
    'billion',
    'billiard',
    'trillion',
    'trilliard'
]

_FRACTION_STRING_DA = {
    2: 'halv',
    3: 'trediedel',
    4: 'fjerdedel',
    5: 'femtedel',
    6: 'sjettedel',
    7: 'syvendedel',
    8: 'ottendedel',
    9: 'niendedel',
    10: 'tiendedel',
    11: 'elftedel',
    12: 'tolvtedel',
    13: 'trettendedel',
    14: 'fjortendedel',
    15: 'femtendedel',
    16: 'sejstendedel',
    17: 'syttendedel',
    18: 'attendedel',
    19: 'nittendedel',
    20: 'tyvendedel'
}

_STRING_FRACTION_DA = invert_dict(_FRACTION_STRING_DA)
_STRING_FRACTION_DA.update({
    'halvdel': 2,
    'kvart': 4
})

_LONG_SCALE = {
    100: 'hundrede',
    1000: 'tusinde',
    1000000: 'million',
    1e9: "milliard",
    1e12: 'billion',
    1e15: "billiard",
    1e18: "trillion",
    1e21: "trilliard",
    1e24: "quadrillion",
    1e27: "quadrilliard"
}

_MULTIPLIER = set(_LONG_SCALE.values())

_STRING_LONG_SCALE = invert_dict(_LONG_SCALE)

# ending manipulation
for number, item in _LONG_SCALE.items():
    if int(number) > 1000:
        name = item + 'er'
        _MULTIPLIER.add(name)
        _STRING_LONG_SCALE[name] = number

_FRACTION_MARKER = set()

_NEGATIVES = {"minus"}

_NUMBER_CONNECTORS = {"og"}

_COMMA = {"komma"}

# Numbers below 1 million are written in one word in Danish, yielding very
# long words
# In some circumstances it may better to seperate individual words
# Set _EXTRA_SPACE_DA=" " for separating numbers below 1 million (
# orthographically incorrect)
# Set _EXTRA_SPACE_DA="" for correct spelling, this is standard

# _EXTRA_SPACE_DA = " "
_EXTRA_SPACE_DA = ""


def is_ordinal_da(input_str):
    """
    This function takes the given text and checks if it is an ordinal number.

    Args:
        input_str (str): the string to check if ordinal
    Returns:
        (bool) or (float): False if not an ordinal, otherwise the number
        corresponding to the ordinal

    ordinals for 1, 3, 7 and 8 are irregular

    only works for ordinals corresponding to the numbers in _DA_NUMBERS

    """

    lowerstr = input_str.lower()

    if lowerstr.startswith("første"):
        return 1
    if lowerstr.startswith("anden"):
        return 2
    if lowerstr.startswith("tredie"):
        return 3
    if lowerstr.startswith("fjerde"):
        return 4
    if lowerstr.startswith("femte"):
        return 5
    if lowerstr.startswith("sjette"):
        return 6
    if lowerstr.startswith("elfte"):
        return 1
    if lowerstr.startswith("tolvfte"):
        return 12

    if lowerstr[-3:] == "nde":
        # from 20 suffix is -ste*
        lowerstr = lowerstr[:-3]
        if lowerstr in _DA_NUMBERS:
            return _DA_NUMBERS[lowerstr]

    if lowerstr[-4:] in ["ende"]:
        lowerstr = lowerstr[:-4]
        if lowerstr in _DA_NUMBERS:
            return _DA_NUMBERS[lowerstr]

    if lowerstr[-2:] == "te":  # below 20 suffix is -te*
        lowerstr = lowerstr[:-2]
        if lowerstr in _DA_NUMBERS:
            return _DA_NUMBERS[lowerstr]

    return False


def nice_number_da(number, speech=True, denominators=range(1, 21)):
    """ Danish helper for nice_number
    This function formats a float to human understandable functions. Like
    4.5 becomes "4 en halv" for speech and "4 1/2" for text
    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3)).replace(".", ",")
    whole, num, den = result
    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)
    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_DA[den]
    if whole == 0:
        if num == 1:
            return_string = '{} {}'.format(num, den_str)
        else:
            return_string = '{} {}e'.format(num, den_str)
    else:
        if num == 1:
            return_string = '{} og {} {}'.format(whole, num, den_str)
        else:
            return_string = '{} og {} {}e'.format(whole, num, den_str)

    return return_string


def pronounce_number_da(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """

    # TODO short_scale, scientific and ordinals
    # currently ignored

    def pronounce_triplet_da(num):
        result = ""
        num = floor(num)
        if num > 99:
            hundreds = floor(num / 100)
            if hundreds > 0:
                if hundreds == 1:
                    result += 'et' + 'hundrede' + _EXTRA_SPACE_DA
                else:
                    result += _NUM_STRING_DA[hundreds] + \
                              'hundrede' + _EXTRA_SPACE_DA
                    num -= hundreds * 100
        if num == 0:
            result += ''  # do nothing
        elif num == 1:
            result += 'et'
        elif num <= 20:
            result += _NUM_STRING_DA[num] + _EXTRA_SPACE_DA
        elif num > 20:
            ones = num % 10
            tens = num - ones
            if ones > 0:
                result += _NUM_STRING_DA[ones] + _EXTRA_SPACE_DA
                if tens > 0:
                    result += 'og' + _EXTRA_SPACE_DA
            if tens > 0:
                result += _NUM_STRING_DA[tens] + _EXTRA_SPACE_DA

        return result

    def pronounce_fractional_da(num, places):
        # fixed number of places even with trailing zeros
        result = ""
        place = 10
        while places > 0:
            # doesn't work with 1.0001 and places = 2: int(
            # number*place) % 10 > 0 and places > 0:
            result += " " + _NUM_STRING_DA[int(num * place) % 10]
            place *= 10
            places -= 1
        return result

    def pronounce_whole_number_da(num, scale_level=0):
        if num == 0:
            return ''

        num = floor(num)
        result = ''
        last_triplet = num % 1000

        if last_triplet == 1:
            if scale_level == 0:
                if result != '':
                    result += '' + 'et'
                else:
                    result += "en"
            elif scale_level == 1:
                result += 'et' + _EXTRA_SPACE_DA + 'tusinde' + _EXTRA_SPACE_DA
            else:
                result += "en " + _NUM_POWERS_OF_TEN[scale_level] + ' '
        elif last_triplet > 1:
            result += pronounce_triplet_da(last_triplet)
            if scale_level == 1:
                result += 'tusinde' + _EXTRA_SPACE_DA
            if scale_level >= 2:
                result += ' ' + _NUM_POWERS_OF_TEN[scale_level]
            if scale_level >= 2:
                result += "er "  # MilliardER, MillioneER

        num = floor(num / 1000)
        scale_level += 1
        return pronounce_whole_number_da(num,
                                         scale_level) + result + _EXTRA_SPACE_DA

    result = ""
    if abs(number) >= 1000000000000000000000000:  # cannot do more than this
        return str(number)
    elif number == 0:
        return str(_NUM_STRING_DA[0])
    elif number < 0:
        return "minus " + pronounce_number_da(abs(number), places)
    else:
        if number == int(number):
            return pronounce_whole_number_da(number)
        else:
            whole_number_part = floor(number)
            fractional_part = number - whole_number_part
            result += pronounce_whole_number_da(whole_number_part)
            if places > 0:
                result += " komma"
                result += pronounce_fractional_da(fractional_part, places)
            return result


def pronounce_ordinal_da(number):
    """
    This function pronounces a number as an ordinal

    1 -> first
    2 -> second

    Args:
        number (int): the number to format
    Returns:
        (str): The pronounced number string.
    """

    # ordinals for 1, 3, 7 and 8 are irregular
    # this produces the base form, it will have to be adapted for genus,
    # casus, numerus

    ordinals = ["nulte", "første", "anden", "tredie", "fjerde", "femte",
                "sjette", "syvende", "ottende", "niende", "tiende"]

    # only for whole positive numbers including zero
    if number < 0 or number != int(number):
        return number
    if number < 10:
        return ordinals[number]
    if number < 30:
        if pronounce_number_da(number)[-1:] == 'e':
            return pronounce_number_da(number) + "nde"
        else:
            return pronounce_number_da(number) + "ende"
    if number < 40:
        return pronounce_number_da(number) + "fte"
    else:
        if pronounce_number_da(number)[-1:] == 'e':
            return pronounce_number_da(number) + "nde"
        else:
            return pronounce_number_da(number) + "ende"


def numbers_to_digits_da(text, short_scale=False,
                               ordinals=False, fractions=True):
    """
    Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numberres should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)
    Returns:
        str
        The original text, with numbers subbed in where appropriate.
    """
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_da(tokens, short_scale, ordinals, fractions)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_da(tokens, short_scale=False,
                                  ordinals=False, fractions=True):
    """
    Extract all numbers from a list of Tokens, with the words that
    represent them.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.

    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = _extract_number_with_text_da(tokens, short_scale, ordinals)

        if not to_replace:
            break

        if isinstance(to_replace.value, float) and not fractions:
            pass
        else:
            results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_da(tokens, short_scale=False, ordinals=False):
    """
    This function extracts a number from a list of Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        ReplaceableNumber

    """
    number, tokens = _extract_number_with_text_da_helper(tokens, short_scale, ordinals)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_da_helper(tokens, short_scale, ordinals):
    """
    Helper for _extract_number_with_text_da.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:
    Returns:
        int or float, [Tokens]
    """
    if ordinals:
        for token in tokens:
            ordinal = is_ordinal_da(token.word)
            if ordinal:
                return ordinal, [token]

    return _extract_real_number_with_text_da(tokens, short_scale)


def _extract_real_number_with_text_da(tokens, short_scale):
    """
    This is handling real numbers.

    Args:
        tokens [Token]:
        short_scale boolean:
    Returns:
        int or float, [Tokens]
        The value parsed, and tokens that it corresponds to.
    """
    number_words = []
    val = _val = _current_val = None
    _comma = False
    to_sum = []

    for idx, token in enumerate(tokens):

        _prev_val = _current_val
        _current_val = None

        word = token.word

        if word in _NUMBER_CONNECTORS and not number_words:
            continue
        if word in (_NEGATIVES | _NUMBER_CONNECTORS | _COMMA):
            number_words.append(token)
            if word in _COMMA:
                _comma = token
                _current_val = _val or _prev_val
            continue

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""

        if word not in _STRING_LONG_SCALE and \
                word not in _DA_NUMBERS and \
                word not in _MULTIPLIER and \
                not is_numeric_da(word) and \
                not is_fractional_da(word):
            words_only = [token.word for token in number_words]
            if _val is not None:
                to_sum.append(_val)
            if to_sum:
                val = sum(to_sum)

            if number_words and (not all([w in _ARTICLES | _NEGATIVES
                                          | _NUMBER_CONNECTORS for w in words_only])
                                 or str(val) == number_words[-1].word):
                break
            else:
                number_words.clear()
                to_sum.clear()
                val = _val = _prev_val = None
            continue
        elif word not in _MULTIPLIER \
                and prev_word not in _MULTIPLIER \
                and prev_word not in _NUMBER_CONNECTORS \
                and prev_word not in _NEGATIVES \
                and prev_word not in _COMMA \
                and prev_word not in _STRING_LONG_SCALE \
                and prev_word not in _DA_NUMBERS \
                and not is_ordinal_da(word) \
                and not is_numeric_da(prev_word) \
                and not is_fractional_da(prev_word):
            number_words = [token]
        else:
            number_words.append(token)

        # is this word already a number or a word of a number?
        _val = _current_val = is_number_da(word)

        # is this a negative number?
        if _current_val is not None and prev_word in _NEGATIVES:
            _val = 0 - _current_val

        # is the prev word a number and should we multiply it?
        if _prev_val is not None and (word in _MULTIPLIER or \
                                      word in _ARTICLES):
            to_sum.append(_prev_val * _current_val or _current_val)
            _val = _current_val = None

        # fraction handling
        _fraction_val = is_fractional_da(word, short_scale=short_scale)
        if _fraction_val:
            if _prev_val is not None and prev_word not in _ARTICLES and \
                    word not in _STRING_FRACTION_DA:  # sammensat brøk
                _val = _prev_val + _fraction_val
                if prev_word not in _NUMBER_CONNECTORS and tokens[idx - 1] not in number_words:
                    number_words.append(tokens[idx - 1])
            elif _prev_val is not None:
                _val = _prev_val * _fraction_val
                if tokens[idx - 1] not in number_words:
                    number_words.append(tokens[idx - 1])
            else:
                _val = _fraction_val
            _current_val = _val

        # directly following numbers without relation
        if (is_numeric_da(prev_word) or prev_word in _DA_NUMBERS) \
                and not _fraction_val and not is_fractional_da(next_word) and not to_sum:
            val = _prev_val
            number_words.pop(-1)
            break

        # TODO: handle spoken time, eg. "kvart over otte" ("quarter past eight"), "fem i ni"...

        # spoken decimals
        if _current_val is not None and _comma:
            # to_sum = [ 1, 0.2, 0.04,...]
            to_sum.append(_current_val if _current_val >= 10 else (_current_val) / (
                                                                          10 ** (token.index - _comma.index)))
            _val = _current_val = None

        if _current_val is not None and next_word in (_NUMBER_CONNECTORS | _COMMA | {""}):
            to_sum.append(_val or _current_val)
            _val = _current_val = None

        if not next_word and number_words:
            val = sum(to_sum) or _val

    return val, number_words


def is_fractional_da(input_str, short_scale=False):
    """
    This function takes the given text and checks if it is a fraction.
    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction
    """
    # account for different numerators, e.g. totrediedele

    input_str = input_str.lower()
    numerator = 1
    prev_number = 0
    denominator = False
    remainder = ""

    # first check if is a fraction containing a char (eg "2/3")
    _bucket = input_str.split('/')
    if look_for_fractions(_bucket):
        numerator = float(_bucket[0])
        denominator = float(_bucket[1])

    if not denominator:
        for fraction in sorted(_STRING_FRACTION_DA.keys(),
                               key=lambda x: len(x),
                               reverse=True):
            if fraction in input_str and not denominator:
                denominator = _STRING_FRACTION_DA.get(fraction)
                remainder = input_str.replace(fraction, "")
                break

        if remainder:
            if not _DA_NUMBERS.get(remainder, False):
                # acount for eineindrittel
                for numstring, number in _DA_NUMBERS.items():
                    if remainder.endswith(numstring):
                        prev_number = _DA_NUMBERS.get(
                            remainder.replace(numstring, "", 1), 0)
                        numerator = number
                        break
                else:
                    return False
            else:
                numerator = _DA_NUMBERS.get(remainder)

    if denominator:
        return prev_number + (numerator / denominator)
    else:
        return False


def is_number_da(word: str):
    if is_numeric_da(word):
        if word.isdigit():
            return int(word)
        else:
            return float(word)
    elif word in _DA_NUMBERS:
        return _DA_NUMBERS.get(word)
    elif word in _STRING_LONG_SCALE:
        return _STRING_LONG_SCALE.get(word)

    return None


def is_numeric_da(input_str):
    """
    Takes in a string and tests to see if it is a number.

    Args:
        text (str): string to test if a number
    Returns:
        (bool): True if a number, else False
    """
    # da float("1.") = 1.0
    if input_str.endswith('.'):
        return False
    try:
        float(input_str)
        return True
    except ValueError:
        return False


def extract_number_da(text, short_scale=False, ordinals=False):
    """
    This function extracts a number from a text string

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found
    """
    numbers = _extract_numbers_with_text_da(tokenize(text.lower()),
                                            short_scale, ordinals)
    # if query ordinals only consider ordinals
    if ordinals:
        numbers = list(filter(lambda x: isinstance(x.value, str)
                                        and x.value.endswith("."),
                              numbers))

    number = numbers[0].value if numbers else None

    if number:
        number = float(number)
        if number.is_integer():
            number = int(number)

    return number



#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from math import floor

from ovos_number_parser.util import (convert_to_mixed_fraction, is_numeric, look_for_fractions, Token)

_NUM_STRING_SV = {
    0: 'noll',
    1: 'en',
    2: 'två',
    3: 'tre',
    4: 'fyra',
    5: 'fem',
    6: 'sex',
    7: 'sju',
    8: 'åtta',
    9: 'nio',
    10: 'tio',
    11: 'elva',
    12: 'tolv',
    13: 'tretton',
    14: 'fjorton',
    15: 'femton',
    16: 'sexton',
    17: 'sjutton',
    18: 'arton',
    19: 'nitton',
    20: 'tjugo',
    30: 'trettio',
    40: 'fyrtio',
    50: 'femtio',
    60: 'sextio',
    70: 'sjuttio',
    80: 'åttio',
    90: 'nittio',
    100: 'hundra'
}

_NUM_POWERS_OF_TEN_SV = [
    'hundra',
    'tusen',
    'miljon',
    'miljard',
    'biljon',
    'biljard',
    'triljon',
    'triljard'
]

_FRACTION_STRING_SV = {
    2: 'halv',
    3: 'tredjedel',
    4: 'fjärdedel',
    5: 'femtedel',
    6: 'sjättedel',
    7: 'sjundedel',
    8: 'åttondel',
    9: 'niondel',
    10: 'tiondel',
    11: 'elftedel',
    12: 'tolftedel',
    13: 'trettondel',
    14: 'fjortondel',
    15: 'femtondel',
    16: 'sextondel',
    17: 'sjuttondel',
    18: 'artondel',
    19: 'nittondel',
    20: 'tjugondel'
}

_EXTRA_SPACE_SV = " "


def nice_number_sv(number, speech=True, denominators=range(1, 21)):
    """ Swedish helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 och en halv" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_SV[den]
    if whole == 0:
        if num == 1:
            return_string = 'en {}'.format(den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1:
        return_string = '{} och en {}'.format(whole, den_str)
    else:
        return_string = '{} och {} {}'.format(whole, num, den_str)
    if num > 1:
        return_string += 'ar'
    return return_string


def pronounce_number_sv(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """

    # TODO short_scale, scientific and ordinals
    # currently ignored

    def pronounce_triplet_sv(num):
        result = ""
        num = floor(num)

        if num > 99:
            hundreds = floor(num / 100)
            if hundreds > 0:
                if hundreds == 1:
                    result += 'ett' + 'hundra'
                else:
                    result += _NUM_STRING_SV[hundreds] + 'hundra'

                num -= hundreds * 100

        if num == 0:
            result += ''  # do nothing
        elif num == 1:
            result += 'ett'
        elif num <= 20:
            result += _NUM_STRING_SV[num]
        elif num > 20:
            tens = num % 10
            ones = num - tens

            if ones > 0:
                result += _NUM_STRING_SV[ones]
            if tens > 0:
                result += _NUM_STRING_SV[tens]

        return result

    def pronounce_fractional_sv(num, places):
        # fixed number of places even with trailing zeros
        result = ""
        place = 10
        while places > 0:
            # doesn't work with 1.0001 and places = 2: int(
            # num*place) % 10 > 0 and places > 0:
            result += " " + _NUM_STRING_SV[int(num * place) % 10]
            place *= 10
            places -= 1
        return result

    def pronounce_whole_number_sv(num, scale_level=0):
        if num == 0:
            return ''

        num = floor(num)
        result = ''
        last_triplet = num % 1000

        if last_triplet == 1:
            if scale_level == 0:
                if result != '':
                    result += '' + 'ett'
                else:
                    result += 'en'
            elif scale_level == 1:
                result += 'ettusen' + _EXTRA_SPACE_SV
            else:
                result += 'en ' + \
                          _NUM_POWERS_OF_TEN_SV[scale_level] + _EXTRA_SPACE_SV
        elif last_triplet > 1:
            result += pronounce_triplet_sv(last_triplet)
            if scale_level == 1:
                result += 'tusen' + _EXTRA_SPACE_SV
            if scale_level >= 2:
                result += _NUM_POWERS_OF_TEN_SV[scale_level]
            if scale_level >= 2:
                result += 'er' + _EXTRA_SPACE_SV  # MiljonER

        num = floor(num / 1000)
        scale_level += 1
        return pronounce_whole_number_sv(num, scale_level) + result

    result = ""
    if abs(number) >= 1000000000000000000000000:  # cannot do more than this
        return str(number)
    elif number == 0:
        return str(_NUM_STRING_SV[0])
    elif number < 0:
        return "minus " + pronounce_number_sv(abs(number), places)
    else:
        if number == int(number):
            return pronounce_whole_number_sv(number)
        else:
            whole_number_part = floor(number)
            fractional_part = number - whole_number_part
            result += pronounce_whole_number_sv(whole_number_part)
            if places > 0:
                result += " komma"
                result += pronounce_fractional_sv(fractional_part, places)
            return result


def pronounce_ordinal_sv(number):
    """
    This function pronounces a number as an ordinal

    1 -> first
    2 -> second

    Args:
        number (int): the number to format
    Returns:
        (str): The pronounced number string.
    """

    # ordinals for 1, 3, 7 and 8 are irregular
    # this produces the base form, it will have to be adapted for genus,
    # casus, numerus

    ordinals = ["noll", "första", "andra", "tredje", "fjärde", "femte",
                "sjätte", "sjunde", "åttonde", "nionde", "tionde"]

    tens = int(floor(number / 10.0)) * 10
    ones = number % 10

    if number < 0 or number != int(number):
        return number
    if number == 0:
        return ordinals[number]

    result = ""
    if number > 10:
        result += pronounce_number_sv(tens).rstrip()

    if ones > 0:
        result += ordinals[ones]
    else:
        result += 'de'

    return result


def _find_numbers_in_text(tokens):
    """Finds duration related numbers in texts and makes a list of mappings.

    The mapping will be for number to token that created it, if no number was
    created from the token the mapping will be from None to the token.

    The function is optimized to generate data that can be parsed to a duration
    so it returns the list in reverse order to make the "size" (minutes/hours/
    etc.) come first and the related numbers afterwards.

    Args:
        tokens: Tokens to parse

    Returns:
        list of (number, token) tuples
    """
    parts = []
    for tok in tokens:
        res = extract_number_sv(tok.word)
        if res:
            parts.insert(0, (res, tok))
            # Special case for quarter of an hour
            if tok.word == 'kvart':
                parts.insert(0, (None, Token('timmar', index=-1)))
        elif tok.word in ['halvtimme', 'halvtimma']:
            parts.insert(0, (30, tok))
            parts.insert(0, (None, Token('minuter', index=-1)))
        else:
            parts.insert(0, (None, tok))
    return parts


def _combine_adjacent_numbers(number_map):
    """Combine adjacent numbers through multiplication.

    Walks through a number map and joins adjasent numbers to handle cases
    such as "en halvtimme" (one half hour).

    Returns:
        (list): simplified number_map
    """
    simplified = []
    skip = False
    for i in range(len(number_map) - 1):
        if skip:
            skip = False
            continue
        if number_map[i][0] and number_map[i + 1][0]:
            combined_number = number_map[i][0] * number_map[i + 1][0]
            combined_tokens = (number_map[i][1], number_map[i + 1][1])
            simplified.append((combined_number, combined_tokens))
            skip = True
        else:
            simplified.append((number_map[i][0], (number_map[i][1],)))

    if not skip:
        simplified.append((number_map[-1][0], (number_map[-1][1],)))
    return simplified


def extract_number_sv(text, short_scale=True, ordinals=False):
    """
    This function prepares the given text for parsing by making
    numbers consistent, getting rid of contractions, etc.
    Args:
        text (str): the string to normalize
    Returns:
        (int) or (float): The value of extracted number
    """
    # TODO: short_scale and ordinals don't do anything here.
    # The parameters are present in the function signature for API
    # compatibility reasons.
    text = text.lower()
    aWords = text.split()
    and_pass = False
    valPreAnd = False
    val = False
    count = 0
    while count < len(aWords):
        word = aWords[count]
        if is_numeric(word):
            val = float(word)
        elif word == "första":
            val = 1
        elif word == "andra":
            val = 2
        elif word == "tredje":
            val = 3
        elif word == "fjärde":
            val = 4
        elif word == "femte":
            val = 5
        elif word == "sjätte":
            val = 6
        elif is_fractional_sv(word):
            val = is_fractional_sv(word)
        else:
            if word == "en":
                val = 1
            if word == "ett":
                val = 1
            elif word == "två":
                val = 2
            elif word == "tre":
                val = 3
            elif word == "fyra":
                val = 4
            elif word == "fem":
                val = 5
            elif word == "sex":
                val = 6
            elif word == "sju":
                val = 7
            elif word == "åtta":
                val = 8
            elif word == "nio":
                val = 9
            elif word == "tio":
                val = 10
            if val:
                if count < (len(aWords) - 1):
                    wordNext = aWords[count + 1]
                else:
                    wordNext = ""
                valNext = is_fractional_sv(wordNext)

                if valNext:
                    val = val * valNext
                    aWords[count + 1] = ""

        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])
            elif and_pass:
                # added to value, quit here
                val = valPreAnd
                break
            else:
                count += 1
                continue

        aWords[count] = ""

        if and_pass:
            aWords[count - 1] = ''  # remove "och"
            val += valPreAnd
        elif count + 1 < len(aWords) and aWords[count + 1] == 'och':
            and_pass = True
            valPreAnd = val
            val = False
            count += 2
            continue
        elif count + 2 < len(aWords) and aWords[count + 2] == 'och':
            and_pass = True
            valPreAnd = val
            val = False
            count += 3
            continue

        break

    return val or False


def is_fractional_sv(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('ars', -3):
        input_str = input_str[:len(input_str) - 3]  # e.g. "femtedelar"
    if input_str.endswith('ar', -2):
        input_str = input_str[:len(input_str) - 2]  # e.g. "femtedelar"
    if input_str.endswith('a', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "halva"
    if input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "halva"

    aFrac = ["hel", "halv", "tredjedel", "fjärdedel", "femtedel", "sjättedel",
             "sjundedel", "åttondel", "niondel", "tiondel", "elftedel",
             "tolftedel"]
    if input_str.lower() in aFrac:
        return 1.0 / (aFrac.index(input_str) + 1)
    if input_str == "kvart":
        return 1.0 / 4
    if input_str == "trekvart":
        return 3.0 / 4

    return False


def normalize_sv(text, remove_articles=True):
    words = text.split()  # this also removed extra spaces
    normalized = ''
    for word in words:
        # Convert numbers into digits, e.g. "two" -> "2"
        if word == 'en':
            word = 'ett'
        textNumbers = ["noll", "ett", "två", "tre", "fyra", "fem", "sex",
                       "sju", "åtta", "nio", "tio", "elva", "tolv",
                       "tretton", "fjorton", "femton", "sexton",
                       "sjutton", "arton", "nitton", "tjugo"]
        if word in textNumbers:
            word = str(textNumbers.index(word))

        normalized += " " + word

    return normalized[1:]  # strip the initial space
